/* FILE: corr.c 
 * 
 * Copyright (c) 2004 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License  
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "config.h"
#include <osivx.h>
#include <zargs.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <string.h>
#include <tiffio.h>
#include <fftw3.h>


#define CORR_EOSIV_STRLEN 4096


typedef struct
{
  unsigned*   mask;
  osivx_args_t*  arglist;

  int         is_zarg;
  int         list_only;
  char*       fft_wisdom;
} corr_packet;

enum {
  CORR_XCORR_ALG,     CORR_PPROC_ALG,
  CORR_INTERP_ALG,    CORR_IMAGE_XSIZE,
  CORR_IMAGE_YSIZE,   CORR_GRID_XSTART,
  CORR_GRID_YSTART,   CORR_GRID_XSPACING,
  CORR_GRID_YSPACING, CORR_GRID_XSIZE,
  CORR_GRID_YSIZE,    CORR_DISP_XOFFSET,
  CORR_DISP_YOFFSET,  CORR_DISP_XMAX,
  CORR_DISP_YMAX,     CORR_DISP_XITER,
  CORR_DISP_YITER,    CORR_WIND_XSIZE,
  CORR_WIND_YSIZE,    CORR_WIND_XSUPER,
  CORR_WIND_YSUPER,   CORR_MOVIE_START,
  CORR_MOVIE_FINISH,  CORR_MOVIE_SKIP,
  CORR_MOVIE_SET,     CORR_NUM_ITER,
  CORR_NUM_THREAD,    CORR_FFT_WISDOM,
  CORR_WRITE_MAP,     CORR_VECTORS_NAME,
  CORR_PARAM_NAME,    CORR_WIND_SIZE,
  CORR_WIND_SUPER,    CORR_LIST_ONLY,
  CORR_NOPTS
}; 

enum {
  CORR_OPT_NONE = 0,
  CORR_OPT_WEAK = 1,
  CORR_OPT_HARD = 2
};
 
static const char* corr_usage =
"Usage: osiv_corr [ -o output ] [ -p file ] movie_a [ movie_b ]\n\n"
"More information is available in the 'osiv_corr' manpage as well\n"
"as in the online documentation.\n";

static unsigned corr_file_nopts = 31;
static zfile_item corr_file_opts[] = {
   {"disp_xiter",     CORR_DISP_XITER},
   {"disp_xmax",      CORR_DISP_XMAX},
   {"disp_xoffset",   CORR_DISP_XOFFSET},
   {"disp_yiter",     CORR_DISP_YITER},
   {"disp_ymax",      CORR_DISP_YMAX},
   {"disp_yoffset",   CORR_DISP_YOFFSET},
   {"fft_wisdom",     CORR_FFT_WISDOM},
   {"grid_xsize",     CORR_GRID_XSIZE},
   {"grid_xspace",    CORR_GRID_XSPACING},
   {"grid_xstart",    CORR_GRID_XSTART},
   {"grid_ysize",     CORR_GRID_YSIZE},
   {"grid_yspace",    CORR_GRID_YSPACING},
   {"grid_ystart",    CORR_GRID_YSTART},
   {"image_xsize",    CORR_IMAGE_XSIZE},
   {"image_ysize",    CORR_IMAGE_YSIZE},
   {"interp_alg",     CORR_INTERP_ALG},
   {"movie_finish",   CORR_MOVIE_FINISH},
   {"movie_set",      CORR_MOVIE_SET},
   {"movie_skip",     CORR_MOVIE_SKIP},
   {"movie_start",    CORR_MOVIE_START},
   {"num_iter",       CORR_NUM_ITER},
   {"num_thread",     CORR_NUM_THREAD},
   {"pproc_alg",      CORR_PPROC_ALG},
   {"wind_size",      CORR_WIND_SIZE},
   {"wind_super",     CORR_WIND_SUPER},
   {"wind_xsize",     CORR_WIND_XSIZE},
   {"wind_xsuper",    CORR_WIND_XSUPER},
   {"wind_ysize",     CORR_WIND_YSIZE},
   {"wind_ysuper",    CORR_WIND_YSUPER},
   {"write_map",      CORR_WRITE_MAP},
   {"xcorr_alg",      CORR_XCORR_ALG},
   {NULL, 0}
};

static unsigned corr_long_nopts = 32; 
static zopts_item corr_long_opts[] = {
   {"disp-xiter",   ZOPTS_WARG,  CORR_DISP_XITER},
   {"disp-xmax",    ZOPTS_WARG,  CORR_DISP_XMAX},
   {"disp-xoffset", ZOPTS_WARG,  CORR_DISP_XOFFSET},
   {"disp-yiter",   ZOPTS_WARG,  CORR_DISP_YITER},
   {"disp-ymax",    ZOPTS_WARG,  CORR_DISP_YMAX},
   {"disp-yoffset", ZOPTS_WARG,  CORR_DISP_YOFFSET},
   {"fft-wisdom",   ZOPTS_WARG,  CORR_FFT_WISDOM},
   {"grid-xsize",   ZOPTS_WARG,  CORR_GRID_XSIZE},
   {"grid-xspace",  ZOPTS_WARG,  CORR_GRID_XSPACING},
   {"grid-xstart",  ZOPTS_WARG,  CORR_GRID_XSTART},
   {"grid-ysize",   ZOPTS_WARG,  CORR_GRID_YSIZE},
   {"grid-yspace",  ZOPTS_WARG,  CORR_GRID_YSPACING},
   {"grid-ystart",  ZOPTS_WARG,  CORR_GRID_YSTART},
   {"image-xsize",  ZOPTS_WARG,  CORR_IMAGE_XSIZE},
   {"image-ysize",  ZOPTS_WARG,  CORR_IMAGE_YSIZE},
   {"interp-alg",   ZOPTS_WARG,  CORR_INTERP_ALG},
   {"list",         ZOPTS_NARG,  CORR_LIST_ONLY},
   {"movie-finish", ZOPTS_WARG,  CORR_MOVIE_FINISH},
   {"movie-set",    ZOPTS_WARG,  CORR_MOVIE_SET},
   {"movie-skip",   ZOPTS_WARG,  CORR_MOVIE_SKIP},
   {"movie-start",  ZOPTS_WARG,  CORR_MOVIE_START},
   {"num-iter",     ZOPTS_WARG,  CORR_NUM_ITER},
   {"num-thread",   ZOPTS_WARG,  CORR_NUM_THREAD},
   {"pproc-alg",    ZOPTS_WARG,  CORR_PPROC_ALG},
   {"wind-size",    ZOPTS_WARG,  CORR_WIND_SIZE},
   {"wind-super",   ZOPTS_WARG,  CORR_WIND_SUPER},
   {"wind-xsize",   ZOPTS_WARG,  CORR_WIND_XSIZE},
   {"wind-xsuper",  ZOPTS_WARG,  CORR_WIND_XSUPER},
   {"wind-ysize",   ZOPTS_WARG,  CORR_WIND_YSIZE},
   {"wind-ysuper",  ZOPTS_WARG,  CORR_WIND_YSUPER},
   {"write-map",    ZOPTS_NARG,  CORR_WRITE_MAP},
   {"xcorr-alg",    ZOPTS_WARG,  CORR_XCORR_ALG},
   {NULL, 0, 0}
};

static unsigned corr_shrt_nopts = 3; 
static zopts_item corr_shrt_opts[] = {
   {"l",           ZOPTS_NARG,  CORR_LIST_ONLY},
   {"o",           ZOPTS_WARG,  CORR_VECTORS_NAME},
   {"p",           ZOPTS_WARG,  CORR_PARAM_NAME},
   {NULL, 0, 0}
};

enum {
  CORR_EMULTD,        CORR_ENINT,
  CORR_EINTER,        CORR_ESIGNAL,
  CORR_ESIGHUP,       CORR_ESIGINT,
  CORR_ESIGQUIT,      CORR_ESIGTERM,
  CORR_ESIGFPE,       CORR_ESIGILL,
  CORR_ESIGSEGV,      CORR_ESIGBUS,
  CORR_ESIGABRT,      CORR_EARGV,
  CORR_EMSIZE,        CORR_EMTYPE,
  CORR_EMOPEN,        CORR_EPNEED,
  CORR_EPWONT,        CORR_EPLOST,
  CORR_EPBUFF,        CORR_EPFILE,
  CORR_EPJUNK,        CORR_EPFORM,
  CORR_EOWISE,        CORR_EBWISE,
  CORR_EMEM,          CORR_EOSIV
};

#define PHONE_HOME \
"\nYou have most likely found a bug within the program.\n"\
"Your help in resolving this problem would be very much\n"\
"appreciated. Please visit the website at www.osiv.org,\n"\
"and submit a bug report which includes information on\n"\
"the situation (and files) with which this occurred.\n"

const char *
corr_strerr(
   int id)
{
   const char* s;

   switch(id)
   {
   case CORR_EMEM:
      s = "unable to allocate memory";
      break;
   case CORR_EPBUFF:
      s = "argument to option `%s\' is too long";
      break;
   case CORR_EPFILE:
      s = "unable to open parameter file: %s";
      break;
   case CORR_EPJUNK:
      s = "parameter file contains binary junk: %s";
      break;
   case CORR_EPFORM:
      s = "invalid parameter file line: %s";
      break;
   case CORR_EOWISE:
      s = "unable to open wisdom file: %s";
      break;
   case CORR_EBWISE:
      s = "invalid fftw wisdom file: %s";
      break;
   case CORR_EPLOST:
      s = "found unrecognized option `%s\'";
      break;
   case CORR_EPWONT:
      s = "option `%s\' doesn't take argument";
      break;
   case CORR_EPNEED:
      s = "option `%s\' requires an argument";
      break;
   case CORR_EARGV:
      s = "requires exactly one or two arguments";
      break;
   case CORR_EMSIZE:
      s = "movie type requires explicit image size";
      break;
   case CORR_EMTYPE:
      s = "movie has unrecogized type: %s";
      break;
   case CORR_EMOPEN:
      s = "unable to open movie file: %s";
      break;
   case CORR_EMULTD:
      s = "option %s defined more than once";
      break;
   case CORR_ENINT:
      s = "option %s expects integer but got `%s\'";
      break;
   case CORR_ESIGNAL:
      s = "unable to install signal handling";
      break;
   case CORR_ESIGINT:
      s = "process caught user interupt";
      break;
   case CORR_ESIGHUP:
      s = "process caught HUP signal";
      break;
   case CORR_ESIGQUIT:
      s = "process caught QUIT signal";
      break;
   case CORR_ESIGTERM:
      s = "process caught TERM signal";
      break;
   case CORR_ESIGFPE:
      s = "caught floating point exception\n" PHONE_HOME;
      break;
   case CORR_ESIGILL:
      s = "caught illegal instruction\n" PHONE_HOME;
      break;
   case CORR_ESIGSEGV:
      s = "caught segmentation fault\n" PHONE_HOME;
      break;
   case CORR_ESIGBUS:
      s = "caught bus error signal\n" PHONE_HOME;
      break;
   case CORR_ESIGABRT:
      s = "caught runtime abort\n" PHONE_HOME;
      break;
   case CORR_EINTER:
      s = "caught internal program error\n" PHONE_HOME;
      break;
   case CORR_EOSIV:
      s = "%s";
      break;
   default:
       return NULL;
   }

   return s;
}

static void
corr_die(
   int id,...)
{
   va_list ap;

   fprintf(stderr,"osiv_corr error: ");
   if (id == CORR_EOSIV) {
      const char* c;
      osivx_err* err_info;
 
      va_start(ap, id);
      c = va_arg(ap, char*);
      err_info = va_arg(ap, osivx_err*);

      fputs(c, stderr);

      switch (err_info->type) {
      case OSIVX_ERR_STRING:
         fprintf(stderr, ": %s",
            err_info->x.str.data);
         break;
      case OSIVX_ERR_INTEG:
         fprintf(stderr, ": %d",
            err_info->x.integ);
         break;
      }

      fputc('\n', stderr);
      va_end(ap);
   } else {
      const char *c;
      c = corr_strerr(id);

      va_start(ap,id);
      vfprintf(stderr, c, ap);
      fputc('\n', stderr);
      va_end(ap);
   }

   exit(1);
}


static void
corr_signal_dohard(
   int signum)
{
   switch (signum)
   {
   case SIGFPE:
      corr_die(CORR_ESIGFPE);
      break;
   case SIGILL:
      corr_die(CORR_ESIGILL);
      break;
   case SIGSEGV:
      corr_die(CORR_ESIGSEGV);
      break;
   case SIGBUS:
      corr_die(CORR_ESIGBUS);
      break;
   case SIGABRT:
      corr_die(CORR_ESIGABRT);
      break;
   default:
      corr_die(CORR_EINTER);
   }
}

static void
corr_signal_dosoft(
   int signum)
{
   switch (signum)
   {  
   case SIGHUP:
      corr_die(CORR_ESIGHUP);
      break;
   case SIGINT:
      corr_die(CORR_ESIGINT);
      break;
   case SIGQUIT:
      corr_die(CORR_ESIGQUIT);
      break;
   case SIGTERM:
      corr_die(CORR_ESIGTERM);
      break;
   default:
      corr_die(CORR_EINTER);
   }
}

static void
corr_signal_mkhard()
{
   struct sigaction newact;
   sigset_t block_mask;

   /* make blocking mask */
   if (sigfillset(&block_mask))
     corr_die(CORR_ESIGNAL);

   /* make action */
   memset(&newact,0,sizeof(newact));
   newact.sa_mask = block_mask;
   newact.sa_handler = &corr_signal_dohard;

   /* install handlers */
   if (sigaction(SIGFPE, &newact, NULL)  ||
       sigaction(SIGILL, &newact, NULL)  ||
       sigaction(SIGSEGV, &newact, NULL) ||
       sigaction(SIGBUS, &newact, NULL)  ||
       sigaction(SIGABRT, &newact, NULL))
     corr_die(CORR_ESIGNAL);
}

static void
corr_signal_mksoft()
{
   struct sigaction newact;
   sigset_t block_mask;

   /* make blocking mask */
   if (sigfillset(&block_mask) ||
       sigdelset(&block_mask, SIGFPE)  ||
       sigdelset(&block_mask, SIGILL)  ||
       sigdelset(&block_mask, SIGSEGV) ||
       sigdelset(&block_mask, SIGBUS)  ||
       sigdelset(&block_mask, SIGABRT))
     corr_die(CORR_ESIGNAL);

   /* make action */
   memset(&newact,0,sizeof(newact));
   newact.sa_mask = block_mask;
   newact.sa_handler = &corr_signal_dosoft;

   /* install handlers */
   if (sigaction(SIGHUP, &newact, NULL)  ||
       sigaction(SIGINT, &newact, NULL)  ||
       sigaction(SIGQUIT, &newact, NULL) ||
       sigaction(SIGTERM, &newact, NULL))
     corr_die(CORR_ESIGNAL);
}

static void
corr_setint(
   unsigned* preset,
   const char* nm,
   const char* val,
   osivx_ioint_t* out,
   int state)
{
   char* endptr = NULL;

   if (*preset == CORR_OPT_HARD)
     corr_die(CORR_EMULTD,nm);

   *out = strtol(val,&endptr,10);
   if (*endptr != '\0' || val==endptr)
     corr_die(CORR_ENINT,nm,val);

   *preset = state; 
}

static void
corr_setbool(
   unsigned* preset,
   const char* nm,
   const char* val,
   osivx_ioint_t* out,
   int state)
{
   if (*preset == CORR_OPT_HARD)
     corr_die(CORR_EMULTD,nm);

   *out = 1;
   *preset = state;

}

static void
corr_setstr(
   unsigned* preset,
   const char* nm,
   const char* val,
   char** out,
   int state)
{
   if (*preset == CORR_OPT_HARD)
     corr_die(CORR_EMULTD,nm);

   *out = strdup(val);
   if (*out == NULL)
     corr_die(CORR_EMEM);

   *preset = state;
}

static void
corr_opt_cb(
   long int evt,
   const char* var,
   const char* val,
   void* data)
{
   corr_packet* m = (corr_packet*)data;
   osivx_args_t* a = m->arglist;
   int state;

   if (m->is_zarg) {
     state = CORR_OPT_HARD;
   } else {
     state = CORR_OPT_WEAK;
   }

   switch (evt)
   {
   case CORR_XCORR_ALG:
      corr_setint(&m->mask[evt], 
        var, val, &a->alg_xcorr,
        state);
      break;
   case CORR_PPROC_ALG:
      corr_setint(&m->mask[evt],
        var, val, &a->alg_pproc,
        state);
      break;
    case CORR_INTERP_ALG:
      corr_setint(&m->mask[evt],
        var, val, &a->alg_interp,
        state);
      break;
   case CORR_IMAGE_XSIZE:
      corr_setint(&m->mask[evt],
        var, val, &a->image_xsize,
        state);
      break;
   case CORR_IMAGE_YSIZE:
      corr_setint(&m->mask[evt],
        var, val, &a->image_ysize,
        state);
      break;
   case CORR_GRID_XSTART:
      corr_setint(&m->mask[evt],
        var, val, &a->grid_xstart,
        state);
      break;
   case CORR_GRID_YSTART:
      corr_setint(&m->mask[evt],
        var, val, &a->grid_ystart,
        state);
      break;
   case CORR_GRID_XSPACING:
      corr_setint(&m->mask[evt],
        var, val, &a->grid_xspacing,
        state);
      break;
   case CORR_GRID_YSPACING:
      corr_setint(&m->mask[evt],
        var, val, &a->grid_yspacing,
        state);
      break;
   case CORR_GRID_XSIZE:
      corr_setint(&m->mask[evt],
        var, val, &a->grid_xsize,
        state);
      break;
   case CORR_GRID_YSIZE:
      corr_setint(&m->mask[evt],
        var, val, &a->grid_ysize,
        state);
      break;
   case CORR_WIND_XSIZE:
      corr_setint(&m->mask[evt],
        var, val, &a->wind_xsize,
        state);
      break;
   case CORR_WIND_YSIZE:
      corr_setint(&m->mask[evt],
        var, val, &a->wind_ysize,
        state);
      break;
   case CORR_DISP_XOFFSET:
      corr_setint(&m->mask[evt],
        var, val, &a->disp_xoffset,
        state);
      break;
   case CORR_DISP_YOFFSET:
      corr_setint(&m->mask[evt],
        var, val, &a->disp_yoffset,
        state);
      break;
   case CORR_DISP_XMAX:
      corr_setint(&m->mask[evt],
        var, val, &a->disp_xmax,
        state);
      break;
   case CORR_DISP_YMAX:
      corr_setint(&m->mask[evt],
        var, val, &a->disp_ymax,
        state);
      break;
   case CORR_DISP_XITER:
      corr_setint(&m->mask[evt],
        var, val, &a->disp_xiter,
        state);
      break;
   case CORR_DISP_YITER:
      corr_setint(&m->mask[evt],
        var, val, &a->disp_yiter,
        state);
      break;
   case CORR_WIND_XSUPER:
      corr_setint(&m->mask[evt],
        var, val, &a->wind_xsuper,
        state);
      break;
   case CORR_WIND_YSUPER:
      corr_setint(&m->mask[evt],
        var, val, &a->wind_ysuper,
        state);
      break;
   case CORR_MOVIE_START:
      corr_setint(&m->mask[evt],
        var, val, &a->movie_start,
        state);
      break;
   case CORR_MOVIE_FINISH:
      corr_setint(&m->mask[evt],
        var, val, &a->movie_finish,
        state);
      break;
   case CORR_MOVIE_SKIP:
      corr_setint(&m->mask[evt],
        var, val, &a->movie_skip,
        state);
      break;
   case CORR_MOVIE_SET:
      corr_setint(&m->mask[evt],
        var, val, &a->movie_set,
        state);
      break;
   case CORR_NUM_ITER:
      corr_setint(&m->mask[evt],
        var, val, &a->num_iter,
        state);
      break;
   case CORR_NUM_THREAD:
      corr_setint(&m->mask[evt],
        var, val, &a->num_thrd,
        state);
      break;
   case CORR_VECTORS_NAME:
      corr_setstr(&m->mask[evt],
        var, val, &a->vectors_name,
        state);
      break;
   case CORR_WRITE_MAP:
      if (m->is_zarg) {
         corr_setbool(&m->mask[evt],
           var, val, &a->write_map,
           state);
      } else {
         corr_setint(&m->mask[evt],
           var, val, &a->write_map,
           state);
      }
      break;
   case CORR_PARAM_NAME:
      m->is_zarg = 0;
      zfile_parse(val, corr_file_opts,
         corr_file_nopts, corr_opt_cb,
         data);
      m->is_zarg = 1;
      break;
   case CORR_LIST_ONLY:
      corr_setbool(&m->mask[evt],
        var, val, &m->list_only,
        state);
      break;
   case CORR_FFT_WISDOM:
      corr_setstr(&m->mask[evt],
        var, val, &m->fft_wisdom,
        state);
      break;
   case CORR_WIND_SIZE:
      corr_setint(&m->mask[CORR_WIND_XSIZE],
        var, val, &a->wind_xsize, state);
      corr_setint(&m->mask[CORR_WIND_YSIZE],
        var, val, &a->wind_ysize, state);
      break;
   case CORR_WIND_SUPER:
      corr_setint(&m->mask[CORR_WIND_XSUPER],
        var, val, &a->wind_xsuper, state);
      corr_setint(&m->mask[CORR_WIND_YSUPER],
        var, val, &a->wind_ysuper, state);
      break;
   }

   if (m->is_zarg) {
      switch (evt)
      {
      case ZOPTS_ENEED:
         corr_die(CORR_EPNEED,var);
      case ZOPTS_EWONT:
         corr_die(CORR_EPWONT,var);
      case ZOPTS_ELOST:
         corr_die(CORR_EPLOST,var);
      case ZOPTS_EBUFF:
         corr_die(CORR_EPBUFF,var);
      }
   } else {
      switch (evt)
      {
      case ZFILE_EOPEN:
         corr_die(CORR_EPFILE,var);
      case ZFILE_ELOST:
         corr_die(CORR_EPLOST,var);
      case ZFILE_EBUFF:
         corr_die(CORR_EPBUFF,var);
      case ZFILE_EFORM:
         corr_die(CORR_EPFORM,var);
      case ZFILE_EJUNK:
         corr_die(CORR_EPJUNK);
      }
   }
}

static void
corr_setimage(
   char* fname,
   osivx_ioint_t s,
   osivx_ioint_t* x,
   osivx_ioint_t* y)
{
   char* suf;

   suf = strrchr(fname, '.');
   if (suf == NULL)
     corr_die(CORR_EMTYPE,fname);

   if (!strcmp(suf, ".tif") ||
       !strcmp(suf, ".tiff"))
   { 
      char* load;
      size_t flen;
      TIFF* tif;
      uint32_t l,w;

      flen = strlen(fname) + 8 * sizeof(osivx_ioint_t) + 1;
      load = (char*) malloc( flen * sizeof(char) );

      if (strrchr(fname, '%')) {
         sprintf(load, fname, s);
      } else {
         strcpy(load, fname);
      }

      tif = TIFFOpen(load, "r");
      if (tif == NULL)
        corr_die(CORR_EMOPEN,load);

      TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
      TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &l);

      if (x != NULL) *x = w;
      if (y != NULL) *y = l;

      TIFFClose(tif);
      free(load);
      return;
   }

   corr_die(CORR_EMSIZE);
}

static void
corr_setdft(
   corr_packet* m)
{
   unsigned* mask = m->mask;
   osivx_args_t* a = m->arglist;

   int fft_xoffset, fft_yoffset;

   if (!mask[CORR_XCORR_ALG]) {
      a->alg_xcorr = 5;
   }
   if (!mask[CORR_PPROC_ALG]) {
      a->alg_pproc = 0;
   }
   if (!mask[CORR_INTERP_ALG]) {
      a->alg_interp = 1;
   }

   if (!mask[CORR_MOVIE_START]) {
      a->movie_start = 0;
   }
   if (!mask[CORR_MOVIE_FINISH]) {
      a->movie_finish = a->movie_start;
   }
   if (!mask[CORR_MOVIE_SKIP]) {
      a->movie_skip = 1;
   }

   {
      osivx_ioint_t* p_image_xsize;
      osivx_ioint_t* p_image_ysize;
   
      if (mask[CORR_IMAGE_XSIZE]) {
         p_image_xsize = NULL;
      } else {
         p_image_xsize = &(a->image_xsize);
      }

      if (mask[CORR_IMAGE_YSIZE]) {
         p_image_ysize = NULL;
      } else {
         p_image_ysize = &(a->image_ysize);
      }

      if (!mask[CORR_IMAGE_XSIZE] ||
          !mask[CORR_IMAGE_YSIZE] )
      {
         corr_setimage(
           a->movie_aname,
           a->movie_start,
           p_image_xsize,
           p_image_ysize);
      }
   }

   if (!mask[CORR_MOVIE_SET]) {
      if (a->movie_bname != NULL) {
         a->movie_set = 0;
      } else {
         a->movie_set = 1;
      }
   }
 
   if (!mask[CORR_WIND_XSIZE]) {
      if (mask[CORR_WIND_XSUPER]) {
         a->wind_xsize = a->wind_xsuper;
      } else {
         a->wind_xsize = 32;
      }
   }
   if (!mask[CORR_WIND_YSIZE]) {
      if (mask[CORR_WIND_YSUPER]) {
         a->wind_ysize = a->wind_ysuper;
      } else {
         a->wind_ysize = 32;
      }
   }
   if (!mask[CORR_WIND_XSUPER]) {
      a->wind_xsuper = 2*a->wind_xsize;
   }
   if (!mask[CORR_WIND_YSUPER]) {
      a->wind_ysuper = 2*a->wind_ysize;
   }

   if (!mask[CORR_DISP_XOFFSET]) {
      a->disp_xoffset = 0;
   }
   if (!mask[CORR_DISP_YOFFSET]) {
      a->disp_yoffset = 0;
   }
   if (!mask[CORR_DISP_XMAX]) {
      a->disp_xmax = (a->wind_xsize-1)/2;
   }
   if (!mask[CORR_DISP_YMAX]) {
      a->disp_ymax = (a->wind_ysize-1)/2;
   }
   if (!mask[CORR_DISP_XITER]) {
      a->disp_xiter = (a->wind_xsize-1)/2;
   }
   if (!mask[CORR_DISP_YITER]) {
      a->disp_yiter = (a->wind_ysize-1)/2;
   }

   fft_xoffset = (a->wind_xsuper - a->wind_xsize)/2;
   fft_yoffset = (a->wind_ysuper - a->wind_ysize)/2;

   /* the xstart/ystart is set to the lowest
    * value that doesn't cause the grid to
    * to extend over the lefthand/upper
    * edge of the first or second image 
    */
   if (!mask[CORR_GRID_XSTART])
   {
      switch(a->alg_xcorr)
      {
      /* this is conservative for algs
       * 1 and 3 but causes less confusion
       * as people expect symmetry here
       */
      case 1: case 2:
      case 3: case 4:
         if (a->disp_xoffset > 0) {
            a->grid_xstart = a->disp_xoffset +
              a->disp_xmax + 1;
         } else {
            a->grid_xstart = a->disp_xmax + 1;
         }
         break;
      case 5:
         if (a->disp_xoffset > 0) {
            a->grid_xstart = fft_xoffset +
              a->disp_xoffset;
         } else {
            a->grid_xstart = fft_xoffset;
         }
         break;
      case 6:
         if (a->disp_xoffset > 0) {
            a->grid_xstart = fft_xoffset +
              a->disp_xoffset + a->disp_xmax;
         } else {
            a->grid_xstart = fft_xoffset +
              a->disp_xmax;
         }
         break;
      case 7: 
         if (a->disp_xoffset > 0) {
            a->grid_xstart = a->disp_xoffset +
              a->disp_xiter;
         } else {
            a->grid_xstart = a->disp_xiter;
         }
         break;
      default:
         if (a->disp_xoffset > 0) {
            a->grid_xstart = a->disp_xmax -
               a->disp_xoffset;
         } else {
            a->grid_xstart = a->disp_xmax;
         }
         break;
      }
   }

   if (!mask[CORR_GRID_YSTART])
   {
      switch(a->alg_xcorr)
      {
      case 1: case 2:
      case 3: case 4:
         if (a->disp_yoffset > 0) {
             a->grid_ystart = a->disp_ymax -
                a->disp_yoffset + 1;
         } else {
             a->grid_ystart = a->disp_ymax + 1;
         }
         break;
      case 5:
         if (a->disp_yoffset > 0) {
            a->grid_ystart = fft_yoffset +
              a->disp_yoffset;
         } else {
            a->grid_ystart = fft_yoffset;
         }
         break;
      case 6:
         if (a->disp_yoffset > 0) {
            a->grid_ystart = fft_yoffset +
              a->disp_yoffset + a->disp_ymax;
         } else {
            a->grid_ystart = fft_yoffset +
              a->disp_ymax;
         }
         break;
      case 7: 
         if (a->disp_yoffset > 0) {
            a->grid_ystart = a->disp_yoffset +
              a->disp_yiter;
         } else {
            a->grid_ystart = a->disp_yiter;
         }
         break;
      default:
         if (a->disp_yoffset > 0) {
            a->grid_ystart = a->disp_ymax -
               a->disp_yoffset;
         } else {
            a->grid_ystart = a->disp_ymax;
         }
         break;
      }
   }

   /* the grid size is set to about 30 vectors
    * by 30 vectors, this is a size that can
    * be evaluated in a reasonable amount of
    * time all algorithms.  we must check if
    * the image can run this many with the
    * minimum spacing.
    */
   if (!mask[CORR_GRID_XSIZE])
   {
      int min_val1, min_val2;

      switch (a->alg_xcorr)
      {
      case 1: case 2:
      case 3: case 4:
         min_val1 = (a->image_xsize - a->grid_xstart +
           a->disp_xoffset - a->disp_xmax - 1) + 1;
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - 1) + 1;
         break;
      case 5:
         min_val1 = (a->image_xsize - a->grid_xstart +
            fft_xoffset + a->disp_xoffset -
            a->wind_xsuper - 1) + 1;
         min_val2 = (a->image_xsize - a->grid_xstart -
            a->wind_xsize - 1) + 1;
         break;
      case 6:
         min_val1 = (a->image_xsize - a->grid_xstart +
           fft_xoffset + a->disp_xoffset -
           a->disp_xmax - a->wind_xsuper - 1) +1;
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - 1) + 1;
         break;
      case 7:
         min_val1 = (a->image_xsize - a->grid_xstart +
           a->disp_xoffset - a->disp_xiter -
           a->wind_xsize - 1)  + 1;
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - 1) + 1;
         break;
      default:
         min_val1 = 100;
         min_val2 = 100;
      }

      min_val1 = 30 < min_val1 ? 30 : min_val1;
      min_val1 = min_val1 < min_val2 ? min_val1 : min_val2;
      a->grid_xsize = min_val1;
   }

   if (!mask[CORR_GRID_YSIZE])
   {
      int min_val1, min_val2;

      switch (a->alg_xcorr)
      {
      case 1: case 2:
      case 3: case 4:
         min_val1 = (a->image_ysize - a->grid_ystart +
           a->disp_yoffset - a->disp_ymax - 1) + 1;
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - 1) +1;
         break;
      case 5:
         min_val1 = (a->image_ysize - a->grid_ystart +
            fft_yoffset + a->disp_yoffset -
            a->wind_ysuper - 1) + 1;
         min_val2 = (a->image_ysize - a->grid_ystart -
            a->wind_ysize - 1) + 1;
         break;
      case 6:
         min_val1 = (a->image_ysize - a->grid_ystart +
           fft_yoffset + a->disp_yoffset -
           a->disp_ymax - a->wind_ysuper - 1) +1;
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - 1) + 1;
         break;
      case 7:
         min_val1 = (a->image_ysize - a->grid_ystart +
           a->disp_yoffset - a->disp_yiter -
           a->wind_ysize - 1)  + 1;
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - 1) + 1;
         break;
      default:
         min_val1 = 100;
         min_val2 = 100;
      }

      min_val1 = 30 < min_val1 ? 30 : min_val1;
      min_val1 = min_val1 < min_val2 ? min_val1 : min_val2;
      a->grid_ysize = min_val1;
   }

   if (!mask[CORR_GRID_XSPACING]) {
      int min_val1, min_val2;

      switch (a->alg_xcorr) {
      case 1: case 2:
      case 3: case 4:
         min_val1 = (a->image_xsize - a->grid_xstart +
           a->disp_xoffset - a->disp_xmax - 1)/(a->grid_xsize-1);
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - a->disp_xmax - 1)/(a->grid_xsize-1);
         break;
      case 5:
         min_val1 = (a->image_xsize - a->grid_xstart +
           fft_xoffset + a->disp_xoffset - a->wind_xsuper - 1)/
           (a->grid_xsize-1);
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - 1)/(a->grid_xsize-1);
         break;
      case 6:
         min_val1 = (a->image_xsize - a->grid_xstart +
           fft_xoffset + a->disp_xoffset -a->disp_xmax -
           a->wind_xsuper - 1)/(a->grid_xsize-1);
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - 1)/(a->grid_xsize-1);
         break;
      case 7:
         min_val1 = (a->image_xsize - a->grid_xstart +
           a->disp_xoffset -a->disp_xiter -
           a->wind_xsize - 1)/(a->grid_xsize-1);
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - 1)/(a->grid_xsize-1);
         break;
      default:
         min_val1 = (a->image_xsize - a->grid_xstart +
           a->disp_xoffset - a->disp_xmax - 1)/(a->grid_xsize-1);
         min_val2 = (a->image_xsize - a->grid_xstart -
           a->wind_xsize - a->disp_xmax - 1)/(a->grid_xsize-1);
         break;
      }

      min_val1 = min_val1 < min_val2 ? min_val1 : min_val2;
      min_val1 = min_val1 < 1 ? 1 : min_val1;
      a->grid_xspacing = min_val1;
   }

   if (!mask[CORR_GRID_YSPACING]) {
      int min_val1, min_val2;

      switch (a->alg_xcorr) {
      case 1: case 2:
      case 3: case 4:
         min_val1 = (a->image_ysize - a->grid_ystart +
           a->disp_yoffset - a->disp_ymax - 1)/(a->grid_ysize-1);
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - a->disp_ymax - 1)/(a->grid_ysize-1);
         break;
      case 5:
         min_val1 = (a->image_ysize - a->grid_ystart +
           fft_yoffset + a->disp_yoffset - a->wind_ysuper - 1)/
           (a->grid_ysize-1);
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - 1)/(a->grid_ysize-1);
         break;
      case 6:
         min_val1 = (a->image_ysize - a->grid_ystart +
           fft_yoffset + a->disp_yoffset -a->disp_ymax -
           a->wind_ysuper - 1)/(a->grid_ysize-1);
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - 1)/(a->grid_ysize-1);
         break;
      case 7:
         min_val1 = (a->image_ysize - a->grid_ystart +
           a->disp_yoffset -a->disp_yiter -
           a->wind_ysize - 1)/(a->grid_ysize-1);
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize  - 1)/(a->grid_ysize-1);
         break;
      default:
         min_val1 = (a->image_ysize - a->grid_ystart +
           a->disp_yoffset - a->disp_ymax - 1)/(a->grid_ysize-1);
         min_val2 = (a->image_ysize - a->grid_ystart -
           a->wind_ysize - a->disp_ymax - 1)/(a->grid_ysize-1);
         break;
      }

      min_val1 = min_val1 < min_val2 ? min_val1 : min_val2;
      min_val1 = min_val1 < 1 ? 1 : min_val1;
      a->grid_yspacing = min_val1;
   }

   if (!mask[CORR_NUM_ITER]) {
      a->num_iter = 3;
   }
   if (!mask[CORR_NUM_THREAD]) {
      a->num_thrd = 2;
   }
   if (!mask[CORR_WRITE_MAP]) {
      a->write_map = 0;
   }

   if (!mask[CORR_VECTORS_NAME]) {
      a->vectors_name = strdup("osiv.out");
   }
   if (!mask[CORR_FFT_WISDOM]) {
      m->fft_wisdom = NULL;
   }
   if (!mask[CORR_LIST_ONLY]) {
      m->list_only = 0;
   }
 
   a->porthole_file = NULL;
}

static void
corr_list(
   osivx_args_t* a)
{
   printf("xcorr_alg = %d\n", a->alg_xcorr);
   printf("pproc_alg = %d\n", a->alg_pproc);
   printf("interp_alg = %d\n", a->alg_interp);
   printf("image_xsize = %d\n", a->image_xsize);
   printf("image_ysize = %d\n", a->image_ysize);
   printf("grid_xstart = %d\n", a->grid_xstart);
   printf("grid_ystart = %d\n", a->grid_ystart);
   printf("grid_xspace = %d\n", a->grid_xspacing);
   printf("grid_yspace = %d\n", a->grid_yspacing);
   printf("grid_xsize = %d\n", a->grid_xsize);
   printf("grid_ysize = %d\n", a->grid_ysize);
   printf("disp_xmax = %d\n", a->disp_xmax);
   printf("disp_ymax = %d\n", a->disp_ymax);
   printf("disp_xoffset = %d\n", a->disp_xoffset);
   printf("disp_yoffset = %d\n", a->disp_yoffset);
   printf("wind_xsize = %d\n", a->wind_xsize);
   printf("wind_ysize = %d\n", a->wind_ysize);
   printf("wind_xsuper = %d\n", a->wind_xsuper);
   printf("wind_ysuper = %d\n", a->wind_ysuper);
   printf("movie_start = %d\n", a->movie_start);
   printf("movie_finish = %d\n", a->movie_finish);
   printf("movie_skip = %d\n", a->movie_skip);
   printf("movie_set = %d\n", a->movie_set);
   printf("num_iter = %d\n", a->num_iter);
   printf("num_thread = %d\n", a->num_thrd);
   printf("write_map = %d\n", a->write_map);
   printf("movie_one = %s\n", a->movie_aname);

   if (a->movie_bname != NULL) {
      printf("movie_two = %s\n", a->movie_bname);
   }

   printf("output = %s\n", a->vectors_name);
   printf("\n");
}

static void
corr_get_wisdom(
   const char* fn)
{
   FILE* fp = fopen(fn, "r");
   if (fp == NULL) {
      corr_die(CORR_EOWISE, fn);
   }

   fftw_import_wisdom_from_file(fp);

   fclose(fp);
}

int
main(int argc, char** argv)
{
   unsigned n,mask[CORR_NOPTS];
   osivx_args_t arglist;
   corr_packet packet;
   osivx_int scode;

   /* install signals */
   corr_signal_mkhard();
   corr_signal_mksoft();

   /* quiet tiff library */
   TIFFSetErrorHandler(NULL);
   TIFFSetWarningHandler(NULL);

   /* null argument state */
   packet.mask = mask;
   memset(mask,0,sizeof(mask));
   packet.arglist = &arglist;

   /* check for usage request */
   if (argc == 1) {
      puts(corr_usage);
      exit(0);
   }

#ifdef DEBUG
   /* verify file variables */
   if (zfile_verify(corr_file_opts,
          corr_file_nopts) != -1) {
      abort();
   }

   /* verify option lists */
   if (zopts_verify(corr_long_opts,
         corr_long_nopts) != -1 ||
       zopts_verify(corr_shrt_opts,
         corr_shrt_nopts) != -1 ) {
      abort();
   }
#endif
  
   /* pull command options */
   packet.is_zarg = 1;
   n = zopts_parse(argc, argv,
     corr_long_opts, corr_long_nopts,
     corr_shrt_opts, corr_shrt_nopts,
     &corr_opt_cb, &packet);

   /* check operands */
   switch (argc-n)
   {
   case 1:
      packet.arglist->movie_aname = argv[n];
      packet.arglist->movie_bname = NULL;
      break;
   case 2:
      packet.arglist->movie_aname = argv[n];
      packet.arglist->movie_bname = argv[n+1];
      break;
   default:
      corr_die(CORR_EARGV);
   }

   /* set defaults when not found */
   corr_setdft(&packet); 

   /* stop if only doing list */
   if (packet.list_only) {
      corr_list(packet.arglist);
      exit(0);
   }

   /* absorb wisdom, if exists */
   if (packet.fft_wisdom) {
      corr_get_wisdom(packet.fft_wisdom);
   }

   /* prepare error info element */
   {
      char* s;

      s  = (char*)malloc(CORR_EOSIV_STRLEN);
      if (s == NULL) {
         corr_die(CORR_EMEM);
      }

      packet.arglist->err_info.x.str.data = s;
      packet.arglist->err_info.x.str.len = CORR_EOSIV_STRLEN;
   }

   /* fire control to library */
   scode = osivx_run(packet.arglist);
   if (scode != 0)
   {
      corr_die(CORR_EOSIV, osivx_strerr(scode),
         &packet.arglist->err_info);
   }

   return 0;
}

