/* FILE: synth.c 
 * 
 * Copyright (c) 2005 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License  
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "config.h"
#include <osivs.h>
#include <zargs.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <string.h>
#include <tiffio.h>


typedef struct
{
  unsigned*    mask;
  osivs_args_t*  arglist;

  int          is_zarg;
  osivs_ioint_t  list_only;
} synth_packet;

enum {
  SYNTH_FNAME0,        SYNTH_FNAME1,
  SYNTH_FILT_SRC,      SYNTH_BACKG_SRC,
  SYNTH_RAND_SEED,     SYNTH_IMAGE_XSIZE,
  SYNTH_IMAGE_YSIZE,   SYNTH_IMAGE_ZSIZE,
  SYNTH_IMAGE_DEPTH,   SYNTH_IMAGE_XEDGE,
  SYNTH_IMAGE_YEDGE,   SYNTH_IMAGE_ZEDGE,
  SYNTH_IMAGE_XYEDGE,  SYNTH_DOT_SPACE,
  SYNTH_DOT_DENSITY,   SYNTH_DOT_SIZE_MIN,
  SYNTH_DOT_SIZE_MAX,  SYNTH_DOT_SIZE_AVG,
  SYNTH_DOT_SIZE_DEV,  SYNTH_DOT_BRIT_MIN,
  SYNTH_DOT_BRIT_MAX,  SYNTH_DOT_BRIT_AVG,
  SYNTH_DOT_BRIT_DEV,  SYNTH_PARAM_NAME,
  SYNTH_LIST_ONLY,     SYNTH_NOPTS 
}; 

enum {
  SYNTH_OPT_NONE = 0,
  SYNTH_OPT_WEAK = 1,
  SYNTH_OPT_HARD = 2
};

static const char* synth_usage =
"Usage: osiv_synth [ ... ] -f formula file1.tif file2.tif \n\n"
"More information is available in the 'osiv_synth' manpage as well\n"
"as in the online documentation.\n";

static unsigned synth_long_nopts = 21; 
static zopts_item synth_long_opts[] = {
   {"bits",         ZOPTS_WARG,  SYNTH_IMAGE_DEPTH},
   {"bright",       ZOPTS_WARG,  SYNTH_DOT_BRIT_AVG},
   {"bright-dev",   ZOPTS_WARG,  SYNTH_DOT_BRIT_DEV},
   {"bright-max",   ZOPTS_WARG,  SYNTH_DOT_BRIT_MAX},
   {"bright-min",   ZOPTS_WARG,  SYNTH_DOT_BRIT_MIN},
   {"density",      ZOPTS_WARG,  SYNTH_DOT_DENSITY},
   {"formula",      ZOPTS_WARG,  SYNTH_FILT_SRC},
   {"height",       ZOPTS_WARG,  SYNTH_IMAGE_YSIZE},
   {"list",         ZOPTS_NARG,  SYNTH_LIST_ONLY},
   {"margin",       ZOPTS_WARG,  SYNTH_IMAGE_XYEDGE},
   {"radius",       ZOPTS_WARG,  SYNTH_DOT_SIZE_AVG},
   {"radius-dev",   ZOPTS_WARG,  SYNTH_DOT_SIZE_DEV},
   {"radius-max",   ZOPTS_WARG,  SYNTH_DOT_SIZE_MAX},
   {"radius-min",   ZOPTS_WARG,  SYNTH_DOT_SIZE_MIN},
   {"seed",         ZOPTS_WARG,  SYNTH_RAND_SEED},
   {"spacing",      ZOPTS_WARG,  SYNTH_DOT_SPACE},
   {"thick",        ZOPTS_WARG,  SYNTH_IMAGE_ZSIZE},
   {"width",        ZOPTS_WARG,  SYNTH_IMAGE_XSIZE},
   {"x-margin",     ZOPTS_WARG,  SYNTH_IMAGE_XEDGE},
   {"y-margin",     ZOPTS_WARG,  SYNTH_IMAGE_YEDGE},
   {"z-margin",     ZOPTS_WARG,  SYNTH_IMAGE_ZEDGE},
   {NULL, 0, 0}
};

static unsigned synth_file_nopts = 20; 
static zfile_item synth_file_opts[] = {
   {"bits",           SYNTH_IMAGE_DEPTH},
   {"bright",         SYNTH_DOT_BRIT_AVG},
   {"bright_dev",     SYNTH_DOT_BRIT_DEV},
   {"bright_max",     SYNTH_DOT_BRIT_MAX},
   {"bright_min",     SYNTH_DOT_BRIT_MIN},
   {"density",        SYNTH_DOT_DENSITY},
   {"formula",        SYNTH_FILT_SRC},
   {"height",         SYNTH_IMAGE_YSIZE},
   {"margin",         SYNTH_IMAGE_XYEDGE},
   {"radius",         SYNTH_DOT_SIZE_AVG},
   {"radius_dev",     SYNTH_DOT_SIZE_DEV},
   {"radius_max",     SYNTH_DOT_SIZE_MAX},
   {"radius_min",     SYNTH_DOT_SIZE_MIN},
   {"seed",           SYNTH_RAND_SEED},
   {"spacing",        SYNTH_DOT_SPACE},
   {"thick",          SYNTH_IMAGE_ZSIZE},
   {"width",          SYNTH_IMAGE_XSIZE},
   {"x_margin",       SYNTH_IMAGE_XEDGE},
   {"y_margin",       SYNTH_IMAGE_YEDGE},
   {"z_margin",       SYNTH_IMAGE_ZEDGE},
   {NULL, 0}
};

static unsigned synth_shrt_nopts = 13; 
static zopts_item synth_shrt_opts[] = {
   {"d",            ZOPTS_WARG,  SYNTH_IMAGE_DEPTH},
   {"f",            ZOPTS_WARG,  SYNTH_FILT_SRC},
   {"g",            ZOPTS_WARG,  SYNTH_IMAGE_XYEDGE},
   {"h",            ZOPTS_WARG,  SYNTH_IMAGE_YSIZE},
   {"l",            ZOPTS_NARG,  SYNTH_LIST_ONLY},
   {"m",            ZOPTS_WARG,  SYNTH_DOT_SPACE},
   {"n",            ZOPTS_WARG,  SYNTH_DOT_DENSITY},
   {"p",            ZOPTS_WARG,  SYNTH_PARAM_NAME},
   {"r",            ZOPTS_WARG,  SYNTH_DOT_SIZE_AVG},
   {"s",            ZOPTS_WARG,  SYNTH_RAND_SEED},
   {"t",            ZOPTS_WARG,  SYNTH_IMAGE_ZSIZE},
   {"w",            ZOPTS_WARG,  SYNTH_IMAGE_XSIZE},
   {"x",            ZOPTS_WARG,  SYNTH_DOT_BRIT_AVG},
   {NULL, 0, 0}
};

enum {
  SYNTH_EMULTD,       SYNTH_ENINT,
  SYNTH_ENFLT,        SYNTH_EINTER,
  SYNTH_ESIGNAL,      SYNTH_ESIGHUP,
  SYNTH_ESIGINT,      SYNTH_ESIGQUIT,
  SYNTH_ESIGTERM,     SYNTH_ESIGFPE,
  SYNTH_ESIGILL,      SYNTH_ESIGSEGV,
  SYNTH_ESIGBUS,      SYNTH_ESIGABRT,
  SYNTH_EARGV,        SYNTH_EPNEED,
  SYNTH_EPWONT,       SYNTH_EPLOST,
  SYNTH_EPBUFF,       SYNTH_EPFILE,
  SYNTH_EPFORM,       SYNTH_EPJUNK,
  SYNTH_EMEM,         SYNTH_EFORM,
  SYNTH_ESOSIV
};

#define PHONE_HOME \
"\nYou have most likely found a bug within the program.\n"\
"Your help in resolving this problem would be very much\n"\
"appreciated. Please visit the website at www.osiv.org,\n"\
"and submit a bug report which includes information on\n"\
"the situation (and files) with which this occurred.\n"

const char *
synth_strerr(
   int id)
{
   const char* s;

   switch(id)
   {
   case SYNTH_EMEM:
      s = "unable to allocate memory";
      break;
   case SYNTH_EPBUFF:
      s = "argument to option `%s\' is too long";
      break;
   case SYNTH_EPLOST:
      s = "found unrecognized option `%s\'";
      break;
   case SYNTH_EPWONT:
      s = "option `%s\' doesn't take argument";
      break;
   case SYNTH_EPNEED:
      s = "option `%s\' requires an argument"; 
      break;
   case SYNTH_EPFILE:
      s = "unable to open parameter file `%s\'";
      break;
   case SYNTH_EPJUNK:
      s = "parameter file `%s\' contains binary junk";
      break;
   case SYNTH_EPFORM:
      s = "invalid parameter file line: %s";
      break;
   case SYNTH_EARGV:
      s = "requires exactly two arguments";
      break;
   case SYNTH_EMULTD:
      s = "option %s defined more than once";
      break;
   case SYNTH_ENINT:
      s = "option %s expects integer but got `%s\'";
      break;
   case SYNTH_ENFLT:
      s = "option %s expects a float but got `%s\'";
      break;
   case SYNTH_ESIGNAL:
      s = "unable to install signal handling";
      break;
   case SYNTH_ESIGINT:
      s = "process caught user interupt";
      break;
   case SYNTH_ESIGHUP:
      s = "process caught HUP signal";
      break;
   case SYNTH_ESIGQUIT:
      s = "process caught QUIT signal";
      break;
   case SYNTH_ESIGTERM:
      s = "process caught TERM signal";
      break;
   case SYNTH_ESIGFPE:
      s = "caught floating point exception\n" PHONE_HOME;
      break;
   case SYNTH_ESIGILL:
      s = "caught illegal instruction\n" PHONE_HOME;
      break;
   case SYNTH_ESIGSEGV:
      s = "caught segmentation fault\n" PHONE_HOME;
      break;
   case SYNTH_ESIGBUS:
      s = "caught bus error signal\n" PHONE_HOME;
      break;
   case SYNTH_ESIGABRT:
      s = "caught runtime abort\n" PHONE_HOME;
      break;
   case SYNTH_EINTER:
      s = "caught internal program error\n" PHONE_HOME;
      break;
   case SYNTH_EFORM:
      s = "displacement formula is required";
      break;
   case SYNTH_ESOSIV:
      s = "%s";
      break;
   default:
      return NULL;
   }

   return s;
}

static void
synth_die(
   int id,...)
{
   va_list ap;
   const char *c;

   fprintf(stderr,"osiv_synth error: ");
   c = synth_strerr(id);

   va_start(ap,id);
   vfprintf(stderr, c, ap);
   va_end(ap);

   fprintf(stderr, "\n");
   exit(1);
}


static void
synth_signal_dohard(
  int signum)
{
   switch (signum)
   {
   case SIGFPE:
      synth_die(SYNTH_ESIGFPE);
      break;
   case SIGILL:
      synth_die(SYNTH_ESIGILL);
      break;
   case SIGSEGV:
      synth_die(SYNTH_ESIGSEGV);
      break;
   case SIGBUS:
      synth_die(SYNTH_ESIGBUS);
      break;
   case SIGABRT:
      synth_die(SYNTH_ESIGABRT);
      break;
   default:
      synth_die(SYNTH_EINTER);
   }
}

static void
synth_signal_dosoft(
   int signum)
{
   switch (signum)
   {  
   case SIGHUP:
      synth_die(SYNTH_ESIGHUP);
      break;
   case SIGINT:
      synth_die(SYNTH_ESIGINT);
      break;
   case SIGQUIT:
      synth_die(SYNTH_ESIGQUIT);
      break;
   case SIGTERM:
      synth_die(SYNTH_ESIGTERM);
      break;
   default:
      synth_die(SYNTH_EINTER);
   }
}

static void
synth_signal_mkhard()
{
   struct sigaction newact;
   sigset_t block_mask;

   /* make blocking mask */
   if (sigfillset(&block_mask))
     synth_die(SYNTH_ESIGNAL);

   /* make action */
   memset(&newact,0,sizeof(newact));
   newact.sa_mask = block_mask;
   newact.sa_handler = &synth_signal_dohard;

   /* install handlers */
   if (sigaction(SIGFPE, &newact, NULL)  ||
       sigaction(SIGILL, &newact, NULL)  ||
       sigaction(SIGSEGV, &newact, NULL) ||
       sigaction(SIGBUS, &newact, NULL)  ||
       sigaction(SIGABRT, &newact, NULL))
     synth_die(SYNTH_ESIGNAL);
}

static void
synth_signal_mksoft()
{
   struct sigaction newact;
   sigset_t block_mask;

   /* make blocking mask */
   if (sigfillset(&block_mask) ||
       sigdelset(&block_mask, SIGFPE)  ||
       sigdelset(&block_mask, SIGILL)  ||
       sigdelset(&block_mask, SIGSEGV) ||
       sigdelset(&block_mask, SIGBUS)  ||
       sigdelset(&block_mask, SIGABRT))
     synth_die(SYNTH_ESIGNAL);

   /* make action */
   memset(&newact,0,sizeof(newact));
   newact.sa_mask = block_mask;
   newact.sa_handler = &synth_signal_dosoft;

   /* install handlers */
   if (sigaction(SIGHUP, &newact, NULL)  ||
       sigaction(SIGINT, &newact, NULL)  ||
       sigaction(SIGQUIT, &newact, NULL) ||
       sigaction(SIGTERM, &newact, NULL))
     synth_die(SYNTH_ESIGNAL);
}

static void
synth_setint(
   unsigned* preset,
   const char* nm,
   const char* val,
   osivs_ioint_t* out,
   int state)
{
   char* endptr = NULL;

   if (*preset == SYNTH_OPT_HARD)
     synth_die(SYNTH_EMULTD,nm);

   *out = strtol(val,&endptr,10);
   if (*endptr != '\0' || val==endptr)
     synth_die(SYNTH_ENINT,nm,val);

   *preset = state; 
}

static void
synth_setflt(
   unsigned* preset,
   const char* nm,
   const char* val,
   osivs_ioreal_t* out,
   int state)
{
   char* endptr = NULL;

   if (*preset == SYNTH_OPT_HARD)
      synth_die(SYNTH_EMULTD,nm);

   *out = strtod(val,&endptr);
   if (*endptr != '\0' || val==endptr)
     synth_die(SYNTH_ENFLT,nm,val);

   *preset = state;
}

/* commented out until it's needed */
static void
synth_setbool(
   unsigned* preset,
   const char* nm,
   const char* val,
   osivs_ioint_t* out,
   int state)
{
   if (*preset == SYNTH_OPT_HARD)
     synth_die(SYNTH_EMULTD,nm);

   *out = 1;
   *preset = state;
}

static void
synth_setstr(
   unsigned* preset,
   const char* nm,
   const char* val,
   char** out,
   int state)
{
   if (*preset == SYNTH_OPT_HARD)
     synth_die(SYNTH_EMULTD,nm);

   *out = strdup(val);
   if (*out == NULL)
     synth_die(SYNTH_EMEM);

   *preset = state;
}

static void
synth_opt_cb(
   long int evt,
   const char* var,
   const char* val,
   void* data)
{
   synth_packet* m = (synth_packet*)data;
   osivs_args_t* a = m->arglist;
   int state;

   if (m->is_zarg) {
     state = SYNTH_OPT_HARD;
   } else {
     state = SYNTH_OPT_WEAK;
   }

   switch (evt)
   {
   case SYNTH_FILT_SRC:
      synth_setstr(&m->mask[evt],
        var, val, &a->filt_src,
        state);
      break;
   case SYNTH_RAND_SEED:
      synth_setint(&m->mask[evt],
        var, val, &a->rand_seed,
        state);
      break;
   case SYNTH_IMAGE_XSIZE:
      synth_setint(&m->mask[evt],
        var, val, &a->image_xsize,
        state);
      break;
   case SYNTH_IMAGE_YSIZE:
      synth_setint(&m->mask[evt],
        var, val, &a->image_ysize,
        state);
      break;
   case SYNTH_IMAGE_ZSIZE:
      synth_setflt(&m->mask[evt],
        var, val, &a->image_zsize,
        state);
      break;
   case SYNTH_IMAGE_DEPTH:
      synth_setint(&m->mask[evt],
        var, val, &a->image_depth,
        state);
      break;
   case SYNTH_IMAGE_XEDGE:
      synth_setflt(&m->mask[evt],
        var, val, &a->image_xedge,
        state);
      break;
   case SYNTH_IMAGE_YEDGE:
      synth_setflt(&m->mask[evt],
        var, val, &a->image_yedge,
        state);
      break;
   case SYNTH_IMAGE_ZEDGE:
      synth_setflt(&m->mask[evt],
        var, val, &a->image_zedge,
        state);
      break;
   case SYNTH_IMAGE_XYEDGE: 
      synth_setflt(&m->mask[evt],
        var, val, &a->image_xedge,
        state);
      synth_setflt(&m->mask[evt],
        var, val, &a->image_yedge,
        state);
      break;
   case SYNTH_DOT_SPACE:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_space,
        state);
      break;
   case SYNTH_DOT_DENSITY:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_density,
        state);
      break;
   case SYNTH_DOT_SIZE_MIN:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_size_min,
        state);
      break;
   case SYNTH_DOT_SIZE_MAX:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_size_max,
        state);
      break;
   case SYNTH_DOT_SIZE_AVG:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_size_avg,
        state);
      break;
   case SYNTH_DOT_SIZE_DEV:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_size_dev,
        state);
      break;
   case SYNTH_DOT_BRIT_MIN:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_brite_min,
        state);
      break;
   case SYNTH_DOT_BRIT_MAX:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_brite_max,
        state);
      break;
   case SYNTH_DOT_BRIT_AVG:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_brite_avg,
        state);
      break;
   case SYNTH_DOT_BRIT_DEV:
      synth_setflt(&m->mask[evt],
        var, val, &a->dot_brite_dev,
        state);
      break;
   case SYNTH_PARAM_NAME:
      m->is_zarg = 0;
      zfile_parse(val, synth_file_opts,
         synth_file_nopts, synth_opt_cb,
         data);
      m->is_zarg = 1;
      break;
   case SYNTH_LIST_ONLY:
      synth_setbool(&m->mask[evt],
        var, val, &m->list_only,
        state);
      break;
   }      

   if (m->is_zarg) {
      switch (evt)
      {
      case ZOPTS_ENEED:
         synth_die(SYNTH_EPNEED,var);
      case ZOPTS_EWONT:
         synth_die(SYNTH_EPWONT,var);
      case ZOPTS_ELOST:
         synth_die(SYNTH_EPLOST,var);
      case ZOPTS_EBUFF:
         synth_die(SYNTH_EPBUFF,var);
      }
   } else {
      switch (evt)
      {
      case ZFILE_EOPEN:
         synth_die(SYNTH_EPFILE,var);
      case ZFILE_ELOST:
         synth_die(SYNTH_EPLOST,var);
      case ZFILE_EBUFF:
         synth_die(SYNTH_EPBUFF,var);
      case ZFILE_EFORM:
         synth_die(SYNTH_EPFORM,var);
      case ZFILE_EJUNK:
         synth_die(SYNTH_EPJUNK);
      }
   }
}

static void
synth_setdft(
   synth_packet* m)
{
   unsigned* mask = m->mask;
   osivs_args_t* a = m->arglist;

   if (!mask[SYNTH_FILT_SRC]) {
      synth_die(SYNTH_EFORM);
   }

   if (!mask[SYNTH_IMAGE_XSIZE]) {
      a->image_xsize = 640;
   }
   if (!mask[SYNTH_IMAGE_YSIZE]) {
      a->image_ysize = 480;
   }
   if (!mask[SYNTH_IMAGE_ZSIZE]) {
      a->image_zsize = 1;
   }
   if (!mask[SYNTH_IMAGE_DEPTH]) {
      a->image_depth = 8;
   }
 
   if (!mask[SYNTH_IMAGE_XYEDGE] &&
       !mask[SYNTH_IMAGE_XEDGE]) {
      a->image_xedge = 10;
   }
   if (!mask[SYNTH_IMAGE_XYEDGE] &&
       !mask[SYNTH_IMAGE_YEDGE]) {
      a->image_yedge = 10;
   }
   if (!mask[SYNTH_IMAGE_ZEDGE]) {
      a->image_zedge = 2;
   }

   if (!mask[SYNTH_RAND_SEED]) {
      a->rand_seed = 110482;
   }
   if (!mask[SYNTH_DOT_DENSITY]) {
      a->dot_density = 0.025;
   }
   if (!mask[SYNTH_DOT_SPACE]) {
      a->dot_space = 0.0;
   }

   if (!mask[SYNTH_DOT_SIZE_AVG]) {
      a->dot_size_avg = 1.5;
   }
   if (!mask[SYNTH_DOT_SIZE_DEV]) {
      a->dot_size_dev = 0.0;
   }
   if (!mask[SYNTH_DOT_SIZE_MIN]) {
      a->dot_size_min = 0.0;
   }
   if (!mask[SYNTH_DOT_SIZE_MAX]) {
      a->dot_size_max = a->dot_size_avg
        + 2 * a->dot_size_dev + 1E-8;
   }

   if (!mask[SYNTH_DOT_BRIT_AVG]) {
      a->dot_brite_avg = 0.75;
   }
   if (!mask[SYNTH_DOT_BRIT_DEV]) {
      a->dot_brite_dev = 0.0;
   }
   if (!mask[SYNTH_DOT_BRIT_MIN]) {
      a->dot_brite_min = 0.0;
   }
   if (!mask[SYNTH_DOT_BRIT_MAX]) {
      a->dot_brite_max = a->dot_brite_avg 
        + 2 * a->dot_brite_dev + 1E-8;
   }

   if (!mask[SYNTH_LIST_ONLY]) {
      m->list_only = 0;
   }
}

static void 
synth_list(
   osivs_args_t* a)
{
   printf("formula = %s\n", a->filt_src);
   printf("seed = %d\n", a->rand_seed); 
   printf("width = %d\n", a->image_xsize);
   printf("height = %d\n", a->image_ysize);
   printf("thick = %g\n", a->image_zsize);
   printf("bits = %d\n", a->image_depth);
   printf("x_margin = %g\n", a->image_xedge);
   printf("y_margin = %g\n", a->image_yedge);
   printf("z_margin = %g\n", a->image_zedge);
   printf("spacing = %g\n", a->dot_space);
   printf("density = %g\n", a->dot_density);
   printf("radius = %g\n", a->dot_size_avg);
   printf("radius_dev = %g\n", a->dot_size_dev);
   printf("radius_min = %g\n", a->dot_size_min);
   printf("radius_max = %g\n", a->dot_size_max);
   printf("bright = %g\n", a->dot_brite_avg);
   printf("bright_dev = %g\n", a->dot_brite_dev);
   printf("bright_min = %g\n", a->dot_brite_min);
   printf("bright_max = %g\n", a->dot_brite_max);
   printf("output1 = %s\n", a->out_fnam0);
   printf("output2 = %s\n", a->out_fnam1);

   printf("\n");
}

int
main(int argc, char** argv)
{
   unsigned n,mask[SYNTH_NOPTS];
   osivs_args_t arglist;
   synth_packet packet;
   osivs_int scode=0;

   /* install signals */
   synth_signal_mkhard();
   synth_signal_mksoft();

   /* quiet tiff library */
   TIFFSetErrorHandler(NULL);
   TIFFSetWarningHandler(NULL);

   /* null argument state */
   packet.mask = mask;
   memset(mask,0,sizeof(mask));
   packet.arglist = &arglist;

   /* check for usage request */
   if (argc == 1) {
      puts(synth_usage);
      exit(0);
   }

#ifdef DEBUG
   /* verify file variables */
   if (zfile_verify(synth_file_opts,
          synth_file_nopts) != -1) {
      abort();
   }


   /* verify option lists */
   if (zopts_verify(synth_long_opts,
         synth_long_nopts) != -1 ||
       zopts_verify(synth_shrt_opts,
         synth_shrt_nopts) != -1 ) {
      abort();
   }
#endif
  
   /* pull command options */
   packet.is_zarg = 1;
   n = zopts_parse(argc, argv,
     synth_long_opts, synth_long_nopts,
     synth_shrt_opts, synth_shrt_nopts,
     &synth_opt_cb, &packet);

   /* check filename operands */
   if (argc - n != 2) {
      synth_die(SYNTH_EARGV);
   }
   packet.arglist->out_fnam0 = argv[n];
   packet.arglist->out_fnam1 = argv[n+1];

   /* set defaults when not found */
   synth_setdft(&packet); 

   /* show arguments if requested */
   if (packet.list_only) {
      synth_list(packet.arglist);
      exit(0);
   }

   /* fire control to library */
   scode = osivs_run(packet.arglist);
   if (scode != 0)
   {
     synth_die(SYNTH_ESOSIV,
       osivs_strerr(scode));
   }

   return 0;
}

