/* FILE: draw.c
 *
 * Copyright (c) 2003-2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivs-devel.h"

#include <string.h>
#include <math.h>


/**********************************************************
 * The following function renders the points which are
 * described in the 'pts' array to the image 'image'.
 * It returns zero on success and an error code on failure.
 **********************************************************/
osivs_int
osivs_draw (
   osivs_opts_t* p,
   osivs_pts_t* pts,
   osivs_real* image,
   osivs_int which)
{
   osivs_int i,j,k;
   osivs_int xl,yl;
   osivs_real maxval;
   osivs_real dscale;
   osivs_real mindist;
   osivs_real* pts_bt;

   /* Useful constants */
   xl = p->image_xsize;
   yl = p->image_ysize;
   maxval = (1<<p->image_depth)-1;

   /* Figure out which bright field */
   pts_bt = which ? pts->bt2 : pts->bt1;

   /* Zero image before we start*/
   memset(image, 0, xl*yl*sizeof(osivs_real));

   /* Traverse points */
   for (i = 0; i < pts->num; i++)
   {
      osivs_real x = pts->x[i];
      osivs_real y = pts->y[i];
      osivs_real z = pts->z[i];
      osivs_real sz = pts->sz[i];
      osivs_real bt = pts_bt[i];
      osivs_int xmin,xmax;
      osivs_int ymin,ymax;

      /* Scale brightness to image */
      bt = maxval * bt; 

      /* Find distance to center of plane */
      z = z - (osivs_real)p->image_zsize / 2;

      /* Find weights of d^2 on particle intensity.
       * Factor of four comes from fact that the
       * std. dev. is equal to 1/2 the size.
       */
      z = (4*z*z)/(2*p->image_zsize*p->image_zsize);
      dscale = 4 / (2 * sz * sz);

      /* Calculate bounding box */
      mindist = sqrt(log(maxval*bt)/dscale);

      xmin = x-mindist<0 ? 0:floor(x-mindist);
      xmax = x+mindist>xl-1 ? xl-1:ceil(x+mindist);
      
      ymin = y-mindist<0 ? 0:floor(y-mindist);
      ymax = y+mindist>yl-1 ? yl-1:ceil(y+mindist);

      /* Add particle to image */
      for (j = ymin; j < ymax; j++)
      {
         osivs_real* ln = image+j*xl;
	 for (k = xmin; k < xmax; k++)
         {
	    osivs_real a,b;

            a = x - k + 0.5;
            b = y - j + 0.5;
            a = a * a + b * b;
            a = a*dscale + z;

            *(ln+k) += bt*osivs_exp(-a);
         }
      }
   }

   /* Threshold Image */
   for (i = 0; i < xl*yl; i++) {
      if (*(image+i) > maxval)
         *(image+i) = maxval;
   }

   return 0;
}

