/* FILE: genr.c
 *
 * Copyright (c) 2003-2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivs-devel.h"


/* define to confirm spacing */
#undef VERIFY_SPACING

typedef struct
{
   osivs_qtarg_t* s_pt;
   osivs_qtarg_t* r_pt;
   osivs_real dist_sq;
} osivs_genr_pkt;


/**********************************************************
 * The following function takes a pointer to a quadtree
 * and checks to see if the target point is less than
 * a specified distance away.
 **********************************************************/
static
int osivs_genr_find(
   const osivs_qtree_t* d_pt,
   void * pkt_void)
{
   /* we could pass each of the dimensions
    * by value, but packet saves stack space
    */
   osivs_genr_pkt* pkt = (osivs_genr_pkt*)pkt_void;
   osivs_qtarg_t* s_pt = pkt->s_pt;
   osivs_qtarg_t* r_pt = pkt->r_pt;

   /* check points, traverse parents */
   if ( !d_pt->is_parent ) {
      osivs_real dx, dy, dz, dt;

      dx = d_pt->u.pt.x - s_pt->x;
      dy = d_pt->u.pt.y - s_pt->y;
      dz = d_pt->u.pt.z - s_pt->z;

      dx = dx / OSIVS_RAND_MAX * r_pt->x;
      dy = dy / OSIVS_RAND_MAX * r_pt->y;
      dz = dz / OSIVS_RAND_MAX * r_pt->z;

      dt = dx*dx + dy*dy + dz*dz;

      if (dt < pkt->dist_sq) {
         return 1;
      }

   } else {
      int i;

      /* recurse to check through children */
      for (i = 0; i < 4 /*quadtree*/; i++) {
         osivs_qtree_t* cc = d_pt->u.ch[i];
         if (cc != NULL) {
            if (osivs_genr_find(cc, pkt_void)) {
               return 1;
            }
         }
      }
   }

   return 0;
}


/**********************************************************
 * The following function verifies that the given points
 * obey the minimum spacing requirement
 **********************************************************/
#ifdef VERIFY_SPACING
static
osivs_int osivs_genr_verify(
   osivs_opts_t* p,
   osivs_pts_t* pts)
{
   osivs_int i, j;
   osivs_real dmax;

   dmax = p->dot_space * p->dot_space;

   for (i = 0; i < pts->num; i++) {
      osivs_real dx, dy, dz, dt;

      for (j = 0; j < i; j++) {
         dx = pts->x[i] - pts->x[j];
         dy = pts->y[i] - pts->y[j];
         dz = pts->z[i] - pts->z[j];
         dt = dx*dx + dy*dy + dz*dz;

         if (dt < dmax) {
            return 1;
         }
      }
   }

   return 0;
}
#endif


/**********************************************************
 * The following function generates a set of particle
 * positions that are dispersed throughout the image as
 * described by the incoming parameters.  It allocates
 * all the memory required for the pts struct.
 **********************************************************/
osivs_int osivs_genr(
   osivs_opts_t* p,
   osivs_pts_t* pts)
{
   osivs_int i, pt_num;
   osivs_qtree_t *qtree;
   osivs_qtarg_t scl, off;
   osivs_int scode = 0;
   osivs_ccp_t dist_qt;
   osivs_real dist_sq;

   /* start off null */
   qtree = NULL;
   pts->x = NULL;
   pts->y = NULL;
   pts->z = NULL;
   pts->u = NULL;
   pts->v = NULL;
   pts->w = NULL;
   pts->sz = NULL;
   pts->bt1 = NULL;
   pts->bt2 = NULL;

   /* calculate scale to image */
   scl.x = p->image_xsize + 2 * p->image_xedge;
   scl.y = p->image_ysize + 2 * p->image_yedge;
   scl.z = p->image_zsize + 2 * p->image_zedge;
   pt_num  = scl.x * scl.y * scl.z * p->dot_density;

   off.x = p->image_xedge;
   off.y = p->image_yedge;
   off.z = p->image_zedge;


   /* calculate spacing in user and qtree units */
   dist_qt = scl.x < scl.y ? scl.x : scl.y;
   dist_qt = p->dot_space * OSIVS_RAND_MAX / dist_qt;
   dist_sq = p->dot_space * p->dot_space;

   /* allocate memory for final data form */
   pts->x = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->y = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->z = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->u = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->v = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->w = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->sz = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->bt1 = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->bt2 = (osivs_real*)malloc(pt_num*sizeof(osivs_real));
   pts->num = pt_num;

   if (pts->x == NULL || pts->y == NULL || pts->z == NULL ||
       pts->u == NULL || pts->v == NULL || pts->w == NULL ||
       pts->sz == NULL || pts->bt1 == NULL || pts->bt2 == NULL)
   {
      scode = OSIVS_EMEM;
      goto finished;
   }

   /* we optimize for the case that no mininum
    * spacing is enforced among particles
    */
   if ( p->dot_space == 0 ) {
      for (i = 0; i < pt_num; i++) {
         pts->x[i] = (osivs_real)osivs_rand() / OSIVS_RAND_MAX 
                         * scl.x - p->image_xedge;
         pts->y[i] = (osivs_real)osivs_rand() / OSIVS_RAND_MAX
                         * scl.y - p->image_yedge;
         pts->z[i] = (osivs_real)osivs_rand() / OSIVS_RAND_MAX
                         * scl.z - p->image_zedge;
      }

      return scode;
   }


   /* check for particle within the minimum distance.
    * we narrow the search substantially by using an
    * quadtree to limit to the smallest quadrant.
    */
   for (i = 0; i < pt_num; i++) {
      int matches, tries=0;
      osivs_qtree_t *rr;
      osivs_genr_pkt pkt;
      osivs_qtarg_t s_pt;

      /* make packet for callback */
      pkt.s_pt = &s_pt; 
      pkt.r_pt = &scl;
      pkt.dist_sq = dist_sq;

      /* reject points within distance,
       * keep looping to get good one
       */
      do {
         if (++tries > OSIVS_MAX_TRIES) {
            scode = OSIVS_ETRIES;
            goto finished;
         }

         /* random numbers span space */
         s_pt.x = osivs_rand();
         s_pt.y = osivs_rand();
         s_pt.z = osivs_rand();

         /* check if point is within distance */
         matches = osivs_qtree_invoke(qtree, &s_pt,
           dist_qt, &osivs_genr_find, (void*)&pkt);

      } while (matches);

      /* use cc to avoid repeat search */
      rr = osivs_qtree_add(qtree, &s_pt);
      if (rr == NULL) {
         scode = OSIVS_EMEM;
         goto finished; 
      }
      qtree = rr;
   }

   /* convert quadtree data into a linear array
    * this also scales and offsets coordinates
    */
   osivs_qtree_convert(qtree, &scl, &off, pts);

#ifdef VERIFY_SPACING
   if (osivs_genr_verify(p, pts)) {
      scode = OSIVS_ETREE;
      goto finished;
   }
#endif

finished:
   /* cleanup on event of error */
   if (scode) {
      if (pts->x) osivs_free(pts->x);
      if (pts->y) osivs_free(pts->y);
      if (pts->z) osivs_free(pts->z);
      if (pts->u) osivs_free(pts->u);
      if (pts->v) osivs_free(pts->v);
      if (pts->w) osivs_free(pts->w);
      if (pts->sz) osivs_free(pts->sz);
      if (pts->bt1) osivs_free(pts->bt1);
      if (pts->bt2) osivs_free(pts->bt2);
   }

   /* free temporary quadtree */
   if (qtree) {
      osivs_qtree_free(qtree);
   }

   return scode;
}

