/* FILE: qtree.c
 * 
 * Copyright (c) 2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivs-devel.h"

#include <string.h>


/*  Define a few simple macros for generating
 *  a location key for each quadtree node.
 */
#define MAX_DEPTH (OSIVS_CCP_HI_BIT+1)
#define KEY_XNUM 1
#define KEY_YNUM 0
#define KEY_XBIT (1<<(KEY_XNUM))
#define KEY_YBIT (1<<(KEY_YNUM))

#define GET_KEY(TREE,N)                                  \
 (((unsigned)((TREE).x & (1<<(N))) >> ((N)-KEY_XNUM)) |  \
  ((unsigned)((TREE).y & (1<<(N))) >> ((N)-KEY_YNUM)))   \

enum {
  SIDE_NORTH,
  SIDE_SOUTH,
  SIDE_EAST,
  SIDE_WEST
};



/**********************************************************
 * The following function allocate a quadtree node.  It
 * returns a pointer on success, and a NULL if no memory
 * is available.
 **********************************************************/
osivs_qtree_t* osivs_qtree_alloc()
{
   osivs_qtree_t* cc = (osivs_qtree_t*)
      osivs_malloc(sizeof(osivs_qtree_t));

   return cc;
}


/**********************************************************
 * The following function frees an entire quadtree, this
 * includes any children that are owned by the tree.  It
 * cannot generate an error.
 **********************************************************/
void osivs_qtree_free(osivs_qtree_t* qtree)
{
   osivs_int i; 

   if (qtree->is_parent) {
      for (i = 0; i < 4 /*quadtree*/; i++) {
         if (qtree->u.ch[i] != NULL) { 
            osivs_qtree_free(qtree->u.ch[i]);
         }
      }
   }

   osivs_free(qtree);
}


/**********************************************************
 * The following function adds a node to the specified
 * quadtree.  It returns a pointer on success and NULL
 * on failure.
 **********************************************************/
osivs_qtree_t* osivs_qtree_add(
   osivs_qtree_t* qtree,
   const osivs_qtarg_t* item)
{
   osivs_qtree_t* rt = qtree;
   osivs_int m, key;

   /* shortcut for null input */
   if (qtree == NULL) {
      qtree = osivs_qtree_alloc();
      if (qtree == NULL) {
         return NULL;
      }

      qtree->is_parent = 0;
      qtree->u.pt.x = item->x;
      qtree->u.pt.y = item->y;
      qtree->u.pt.z = item->z;

      return qtree;
   }

   /* start from last bit */
   m = OSIVS_CCP_HI_BIT;


   /* decend into the quadtree */
decend_tree:
   do {
      osivs_qtree_t* cc;

      key = GET_KEY(*item, m);

      if ( !qtree->is_parent ) {
         goto hit_point;
      } 

      cc = qtree->u.ch[key];
      if (cc == NULL) {
         goto hit_blank;
      }

      qtree = cc;
      m--;
   } while (m >= 0);

   /* got through tree without miss
    * exactly matches existing point
    * (absurdly rare event)
    */
   return rt;


   /* found point in proper quad,
    * add to the quadtree
    */
hit_point:
   {
      osivs_qtree_t *rr;
      osivs_int key2;

      /* copy data to new point */
      rr = osivs_qtree_alloc();
      if (rr == NULL) {
         return NULL;
      }

      rr->is_parent = 0;
      rr->u.pt.x = qtree->u.pt.x;
      rr->u.pt.y = qtree->u.pt.y;
      rr->u.pt.z = qtree->u.pt.z;

      /* transform to parent */
      qtree->is_parent = 1;
      memset(qtree->u.ch, 0, sizeof(qtree->u.ch));

      /* place new node as child */
      key2 = GET_KEY(rr->u.pt, m);
      qtree->u.ch[key2] = rr;

      /* if the kyes are different
       * we just copy otherwise we
       * perform a tail recursion
       */
      if (key != key2) {
         goto hit_blank;
      } else {
         goto decend_tree;
      }
   }


hit_blank:
   /* found spot for this point,
    * add to the quadtree
    */
   {
      osivs_qtree_t* cc;

      /* make node for new point */
      cc = osivs_qtree_alloc();
      if (cc == NULL) {
         return NULL;
      }

      cc->is_parent = 0;
      cc->u.pt.x = item->x;
      cc->u.pt.y = item->y;
      cc->u.pt.z = item->z;

      qtree->u.ch[key] = cc;
   }

   return rt;
}


/**********************************************************
 * This function weaves it's way down through the quadtree
 * to construct a stack to the desired depth. It returns
 * the achieved depth.
 **********************************************************/
static
osivs_int osivs_qtree_mk_stack(
   const osivs_qtree_t* qtree,
   const osivs_qtarg_t* item,
   const osivs_qtree_t* qstack[],
   unsigned int kstack[],
   unsigned int depth)
{
   int i, m;

   /* start dropping down tree */
   m = OSIVS_CCP_HI_BIT;
   for (i = 0; i < depth; i++) {
      osivs_qtree_t* cc;
      unsigned key;

      qstack[i] = qtree;
      key = GET_KEY(*item, m);
      kstack[i] = key;

      if ( !qtree->is_parent ) {
         return i + 1;
      }

      cc = qtree->u.ch[key];
      if (cc == NULL) {
         return i + 1;
      }

      qtree = cc;
      m--;
   }

   return depth;
}


/**********************************************************
 * This function returns a pointer to a neighbor of the
 * given quadrant on the given side at the given depth.
 **********************************************************/
static
const osivs_qtree_t* osivs_qtree_neighbor(
   const osivs_qtree_t* qstack[],
   const unsigned int kstack[],
   unsigned int cpos,
   unsigned int which)
{
   const osivs_qtree_t* qtree;
   osivs_qtree_t *cc;
   int i = cpos;

   switch (which) {
   case SIDE_NORTH:
      /* back out of the quadrant */
      while (i >= 0 && !(kstack[i] & KEY_YBIT)) {
         i--;
      }

      if (i < 0) {
         return NULL;
      }

      /* drop down to the gcd neighbor */
      {
         unsigned key = kstack[i];
         qtree = qstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key & ~KEY_YBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
         i++;
      }

      /* dive back into quadrant */
      for (; i < cpos; i++) {
         unsigned key = kstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key | KEY_YBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
      }

      break;
   case SIDE_SOUTH:
      /* back out of the quadrant */
      while (i >= 0 && (kstack[i] & KEY_YBIT)) {
         i--;
      }

      if (i < 0) {
         return NULL;
      }

      /* drop down to the gcd neighbor */
      {
         unsigned key = kstack[i];
         qtree = qstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key | KEY_YBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
         i++;
      }

      /* dive back into quadrant */
      for (; i < cpos; i++) {
         unsigned key = kstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key & ~KEY_YBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
      }

      break;
   case SIDE_WEST:
      /* back out of the quadrant */
      while (i >= 0 && !(kstack [i] & KEY_XBIT)) {
         i--;
      }

      if (i < 0) {
         return NULL;
      }

      /* drop down to the gcd neighbor */
      {
         unsigned key = kstack[i];
         qtree = qstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key & ~KEY_XBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
         i++;
      }

      /* dive back into quadrant */
      for (; i < cpos; i++) {
         unsigned key = kstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key | KEY_XBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
      }

      break;
   case SIDE_EAST:
      /* back out of the quadrant */
      while (i >= 0 && (kstack[i] & KEY_XBIT)) {
         i--;
      }

      if (i < 0) {
         return NULL;
      }

      /* drop down to the gcd neighbor */
      {
         unsigned key = kstack[i];
         qtree = qstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key | KEY_XBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
         i++;
      }

      /* dive back into quadrant */
      for (; i < cpos; i++) {
         unsigned key = kstack[i];

         if ( !qtree->is_parent ) {
            return qtree;
         }

         cc = qtree->u.ch[key & ~KEY_XBIT];
         if (cc == NULL) {
            return NULL;
         }

         qtree = cc;
      }

      break;

   default:
      /* Should never get here */
      cc = NULL;
   }

   return cc;
}


/**********************************************************
 * This function finds the quadrant which is diagonal to
 * the provided quadrant, but is not not a sibling the
 * provided quadrant. In other words, it finds the "far"
 * diagonal quadrant.
 **********************************************************/
static
const osivs_qtree_t* osivs_qtree_diagonal(
   const osivs_qtree_t* qstack[],
   const unsigned kstack[],
   unsigned int cpos)
{
   const osivs_qtree_t* qtree;
   int i, init, xpos, ypos;
   const osivs_qtree_t* cc=NULL;

   i = cpos;
   init = kstack[i--];

   /* back out of the quadrant, noting
    * when either the X or Y bit flips
    */
   for (; i >= 0; i--) {
      int m = init ^ kstack[i];

      if (m & KEY_XBIT) {
         xpos = i;
         goto seek_only_y;
      }

      if (m & KEY_YBIT) {
         ypos = i;
         goto seek_only_x;
      }
   } 

   /* quadrant abutes edge */
   return NULL;
 

   /* back out of the quadrant, noting
    * only when the X bit flips
    */
seek_only_x:
   for (; i >= 0; i--) {
      int m = init ^ kstack[i];

      if (m & KEY_XBIT) {
         xpos = i;
         goto x_lte_y;
      }
   }

   /* quadrant abutes edge */
   return NULL;


   /* back out of the quadrant, noting
    * only when the Y bit flips
    */
seek_only_y:
   for (; i >= 0; i--) {
      int m = init ^ kstack[i];

      if (m & KEY_YBIT) {
         ypos = i;
         goto y_lte_x;
      }
   }

   /* quadrant abutes edge */
   return NULL;


   /* dive back into the quadtree.
    * assume that xpos is less than
    * or equal to ypos.
    */
x_lte_y:
   qtree = qstack[i];
   for (; i < ypos; i++) {
      unsigned key = kstack[i];

      if ( !qtree->is_parent ) {
         return qtree;
      }

      cc = qtree->u.ch[key ^ KEY_XBIT];
      if (cc == NULL) {
         return NULL;
      }
      qtree = cc;
   }

   for (; i <= cpos; i++) {
      unsigned key = kstack[i];

      if ( !qtree->is_parent ) {
         return qtree;
      }

      key ^= KEY_XBIT | KEY_YBIT;
      cc = qtree->u.ch[key];
      if (cc == NULL) {
         return NULL;
      }
      qtree = cc;
   }

   return cc;

   /* dive back into the quadtree.
    * assume that ypos is less than
    * or equal to xpos.
    */
y_lte_x:
   qtree = qstack[i];
   for (; i < xpos; i++) {
      unsigned key = kstack[i];

      if ( !qtree->is_parent ) {
         return qtree;
      }

      cc = qtree->u.ch[key ^ KEY_YBIT];
      if (cc == NULL) {
         return NULL;
      }
      qtree = cc;
   }

   for (; i <= cpos; i++) {
      unsigned key = kstack[i];

      if ( !qtree->is_parent ) {
         return qtree;
      }

      key ^= KEY_XBIT | KEY_YBIT;
      cc = qtree->u.ch[key];
      if (cc == NULL) {
         return NULL;
      }
      qtree = cc;
   }

   return cc;
}


/**********************************************************
 * This function is how most callers will access points
 * contained in the quadtree. It traverses the given
 * quadtree to find the quadrant containing the given
 * target point.  It uses the given distance to calculate
 * how many levels down in the tree it should go; the
 * shortest edge of the invoked quadrant will be longer
 * than this distance.  It then calls the provided
 * callback on the found quadrant and it's immediate
 * neighbors.  The value udata is passed to the callback
 * and can be used to store whatever data is needed by
 * the callback.  this function returns the value of the
 * first callback to return a non-zero value or zero if
 * all return zero.  traversal of neighbors stops at the
 * first non-zero callback.
 **********************************************************/
int osivs_qtree_invoke(
   const osivs_qtree_t* qtree,
   const osivs_qtarg_t* item,
   unsigned distance,
   osivs_qtree_cb_t cb,
   void* udata)
{
   const osivs_qtree_t* qstack[MAX_DEPTH];
   unsigned int kstack[MAX_DEPTH];
   int key, cpos, level;

   /* shortcut for no quadtree */
   if (qtree == NULL) return 0;

   /* max depth to dig */ 
   level = MAX_DEPTH;
   while (distance) {
      distance >>= 1;
      level--;
   }

   if (level < 0) {
      level = 0;
   }

   /* dive in to construct stack */
   cpos = osivs_qtree_mk_stack(qtree,
      item, qstack, kstack, level);

   /* shortcut for no depth */
   if (cpos == 0) {
      int retval = (*cb)(qtree, udata);
      return retval;
   }

   cpos--; /* length -> index */
   key = kstack[cpos];

   /* engage same-quadrant neighbors */
   {
      int retval;

      retval = (*cb)(qstack[cpos], udata);
      if (retval != 0) {
         return retval;
      }
   }

   /* engage x-side neighbors */
   {
      const osivs_qtree_t *cc, *rr;
      int direc, flip;

      if (key & KEY_XBIT) {
         direc = SIDE_EAST;
      } else {
         direc = SIDE_WEST;
      }
      flip = key ^ KEY_XBIT;

      cc = osivs_qtree_neighbor(qstack,
         kstack, cpos, direc);

      if (cc != NULL) {
         if (cc->is_parent) {
            /* act on entire x-side */
            rr = cc->u.ch[flip | KEY_YBIT];
            if (rr != NULL) {
               int retval;

               retval = (*cb)(rr, udata);
               if (retval != 0) {
                  return retval;
               }
            }

            rr = cc->u.ch[flip & ~KEY_YBIT];
            if (rr != NULL) {
               int retval;

               retval = (*cb)(rr, udata);
               if (retval != 0) {
                  return retval;
               }
            }
 
         } else {
            int retval;

            retval = (*cb)(cc, udata);
            if (retval != 0) {
               return retval;
            }
         }
      }
   }
 
   /* engage y-side neighbors */
   {
      const osivs_qtree_t *cc, *rr;
      int direc, flip;

      if (key & KEY_YBIT) {
         direc = SIDE_SOUTH;
      } else {
         direc = SIDE_NORTH;
      }
      flip = key ^ KEY_YBIT;

      cc = osivs_qtree_neighbor(qstack,
         kstack, cpos, direc);

      if (cc != NULL) {
         if (cc->is_parent) {

            /* act on entire y-side */
            rr = cc->u.ch[flip | KEY_XBIT];
            if (rr != NULL) {
               int retval;

               retval = (*cb)(rr, udata);
               if (retval != 0) {
                  return retval;
               }
            }

            rr = cc->u.ch[flip & ~KEY_XBIT];
            if (rr != NULL) {
               int retval;

               retval = (*cb)(rr, udata);
               if (retval != 0) {
                  return retval;
               }
            }
         } else {
            int retval;

            retval = (*cb)(cc, udata);
            if (retval != 0) {
               return retval;
            }
         }
      }
   }
 
   /* engage diag-side neighbors */
   {
      const osivs_qtree_t *cc;

      /* diagonal digs one level deeper
       * than neighbor, so we don't need
       * to get the closest quadrant here
       */

      cc = osivs_qtree_diagonal(qstack,
         kstack, cpos);

      if (cc != NULL) {
         int retval;

         retval = (*cb)(cc, udata);
         if (retval != 0) {
            return retval;
         }
      } 
   }

   return 0;
}


/**********************************************************
 * The following function converts a quadtree into a
 * simple linear array of points.
 **********************************************************/
void
osivs_qtree_convert(
   const osivs_qtree_t* qtree,
   const osivs_qtarg_t* scl,
   const osivs_qtarg_t* off,
   osivs_pts_t* pts)
{
   const osivs_qtree_t* stack[4*MAX_DEPTH];
   osivs_int i=0, j=0, k;

   /* first stack entry is arg */
   stack[i] = qtree;

   /* decend into the stack */
   while (i >= 0) {
      const osivs_qtree_t* cc = stack[i];

      if ( !cc->is_parent ) {
         /* points append themselves to list */
         pts->x[j] = (osivs_real)cc->u.pt.x / OSIVS_RAND_MAX
                         * scl->x - off->x;
         pts->y[j] = (osivs_real)cc->u.pt.y / OSIVS_RAND_MAX
                         * scl->y - off->y;
         pts->z[j] = (osivs_real)cc->u.pt.z / OSIVS_RAND_MAX
                         * scl->z - off->z;
         j++;
      } else {
         /* parents push children onto stack */
         for (k = 0; k < 4 /* quadtree */; k++) {
            if (cc->u.ch[k] != NULL) {
               stack[i] = cc->u.ch[k];
               i++;
            }
         }
      }

      i--;
   }
}

