/* FILE: rand.c
 * 
 * Copyright (c) 2003-2005 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* This algorithm is based on ran1 from Numerical Recipes */

#include "osivs-devel.h"

#include <math.h>

#define NTAB 32
#define IA 16807
#define IQ 127773
#define IM 2147483647
#define IR 2836
#define NDIV (1+(IM-1)/NTAB)


/**********************************************************
 * The following function returns a 32-bit pseudo-random
 * number.  If it is passed a non-zero argument then that
 * integer is used as a seed for the random number
 * generator.
 **********************************************************/
static int32_t osivs_rng(int32_t s)
{
   static int32_t x = 0;
   static int32_t iy = 0;
   static int32_t tbl[NTAB];
  
   if (s || !iy) {
      int i;

      if (!s) s = 1;

      for (i = 7; i >= 0; i--) {
         int32_t h = s/IQ;
         s = IA*(s-h*IQ)-h*IR;
         if (s < 0) s += IM;
      }

      for (i = NTAB-1; i >= 0; i--) {
         int32_t h = s/IQ;
         s = IA*(s-h*IQ)-h*IR;
         if (s < 0) s += IM;
         tbl[i] = s;
      }
      
      x = s;
      iy = s;
   }

   {
      int32_t h = x/IQ;
      x = IA*(x-h*IQ)-h*IR;
      if (x < 0) x += IM;

      h = iy/NDIV;
      iy = tbl[h];
      tbl[h] = x;
   }

   return iy;
}


/**********************************************************
 * This function sets the seed number for the random
 * number generator.  A non-zero number must be specified
 * for reasonable random number.
 **********************************************************/
void osivs_srand(int32_t s)
{
   osivs_rng(s < 0 ? -s : s);
}


/**********************************************************
 * This function returns a psuedo-random integer between
 * zero and OSIVS_RAND_MAX.  This function reads and writes
 * static memory, and so must not be accessed from multiple
 * threads simultaneously.
 **********************************************************/
uint32_t osivs_rand()
{
   return osivs_rng(0);
}


/**********************************************************
 * This function returns a pseudo-random floating-point
 * number between zero and one.  This function relies on
 * osivs_rand() which reads and writes static memory, and
 * so this function must not be accessed from multiple
 * threads simultaneously.
 **********************************************************/
float osivs_randf()
{
   return ((float)osivs_rng(0) + 1)
     / OSIVS_RAND_MAX;
}


/**********************************************************
 * This function returns a normally-distributed random
 * floating-point number with mean of zero and variance
 * of one.  This number may be trivially scaled to any
 * other mean or variance.  The algorithm used by this
 * routine is a Box-Muller transform of the random numbers
 * produced by osivs_randf().
 **********************************************************/
float osivs_randn()
{
   static float store = 0.0;
   static int has_store = 0;

   if ( has_store ) {
      float tmp = store;
      has_store = 0;
      return tmp;
   } else {
      float w,x1,x2;
      do {
         x1 = 2.0*osivs_randf() - 1.0;
         x2 = 2.0*osivs_randf() - 1.0;
         w = x1*x1 + x2*x2;
      } while (w >= 1.0);

      w = sqrt((-2.0*log(w))/w);
      store = x1*w;
      return x2*w;
   }
}

