/* FILE: alg1.c 
 *
 * Copyright (c) 2002-2006 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"


/**********************************************************
 * The following function traverses the passed arguments
 * and returns a non-zero error code if they are invalid.
 **********************************************************/
osivx_int
osivx_alg1_checkargs(
   osivx_args_t * a)
{
   /* check for grid extending over first image */
   if (a->grid_xstart - a->disp_xoffset - a->disp_xmax < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart - a->disp_xoffset + a->disp_xmax +
        (a->grid_xsize - 1) * a->grid_xspacing +
        a->wind_xsize >= a->image_xsize) {
      return OSIVX_EXGRID;
   }
        
   if (a->grid_ystart - a->disp_yoffset - a->disp_ymax < 0) {
      return OSIVX_EYGRID;
   }
   if (a->grid_ystart - a->disp_yoffset + a->disp_ymax +
        (a->grid_ysize - 1) * a->grid_yspacing +
        a->wind_ysize >= a->image_ysize) {
      return OSIVX_EYGRID;
   }

   /* check for grid extending over second image */
   if (a->grid_xstart < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart + a->wind_xsize + (a->grid_xsize - 1)
         * a->grid_xspacing >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart < 0) {
      return OSIVX_EYGRID;
   }
   if (a->grid_ystart + a->wind_ysize + (a->grid_ysize - 1)
         * a->grid_yspacing >= a->image_ysize) {
      return OSIVX_EYGRID;
   }

   return 0;
}


/**********************************************************
 * The following function sums the squared difference 
 * between a window that begins at im0 and a window
 * that begins at im1.  It cannot generate an error.
 **********************************************************/
void
osivx_alg1_sumwind(
   const osivx_imcp_t * im0,
   const osivx_imcp_t * im1,
   osivx_int image_xsize,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_immw_t * map)
{
   osivx_int i, j;
   osivx_immw_t s = 0;

   for (i = wind_ysize - 1; i >= 0; i--)
   {
      const osivx_imcp_t *i0 = im0 + i * image_xsize;
      const osivx_imcp_t *i1 = im1 + i * image_xsize;

      for (j = wind_xsize - 1; j >= 0; j--)
      {
         osivx_imsp_t d = *(i0 + j) - *(i1 + j);
         s += d * d;
      }
   }

   *map = s;
}


/**********************************************************
 * The following function calculates the summed, squared
 * difference between windows for a range of displacements.
 * The regions of interest begin at im0 and im1, and are
 * assumed to have memory at both positive and negative
 * displacements sufficient for the given displacements.
 * It cannot generate an error.
 **********************************************************/
static void
osivx_alg1_getmap(
   const osivx_opts_t * p,
   const osivx_imcp_t * im0,
   const osivx_imcp_t * im1,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_immw_t * map,
   osivx_int * ind)
{
   osivx_int i, j;
   osivx_int z;

   /* offset image to new coords */
   im0 -= p->disp_ymax * p->image_xsize + p->disp_xmax;

   /* set default minimum */
   *ind = z = 0;

   /* traverse map */
   for (i = p->map_ysize - 1; i >= 0; i--)
   {
      const osivx_imcp_t *i0 = im0 + i * p->image_xsize;

      for (j = p->map_xsize - 1; j >= 0; j--)
      {
         osivx_alg1_sumwind(i0 + j, im1,
           p->image_xsize,
           wind_xsize, wind_ysize,
           map + z);

         if (*(map + z) <= *(map + *ind)) {
           *ind = z;
         }

         z++;
      }
   }
}


/**********************************************************
 * The following function converts the map type to a
 * input/output data type.  It cannot generate an error.
 **********************************************************/
static void
osivx_alg1_map2std(
   const osivx_opts_t * p,
   osivx_ioreal_t * dest,
   osivx_immw_t * src)
{
   osivx_int i;

   for (i = p->map_len - 1; i >= 0; i--) {
      *(dest + i) = (osivx_ioreal_t) *(src + i);
   }
}


/**********************************************************
 * The following function is called to initialize the
 * algorithm 1 data.  It returns zero on success and a
 * non-zero number on error.
 **********************************************************/
static osivx_int
osivx_alg1_do_init(
   osivx_wind_alg_t* alg,
   const osivx_opts_t* p,
   const osivx_wind_alg_t* master)
{
   alg->udata[0] = (osivx_immw_t*) osivx_malloc(
      p->map_len * sizeof(osivx_immw_t));
   if (alg->udata[0] == NULL) return OSIVX_EMEM;

   /* quiet compiler warnings */
   master = NULL;

   return 0;
}


/**********************************************************
 * The following function calculates displacement
 * vectors for iter_num windows.  It writes the x
 * displacement to *std_vec, and the y displacement
 * to *(std_vec+1).  If std_map is non-null, then
 * the correlation map is copied to that memory.
 * It returns zero on success and a non-zero number
 * on error.
 **********************************************************/
static osivx_int
osivx_alg1_do_window(
   osivx_wind_alg_t* alg,
   const osivx_opts_t* p,
   osivx_grid_iter_t* iter,
   const osivx_imcp_t* im0,
   const osivx_imcp_t* im1,
   osivx_ioreal_t* std_vec,
   osivx_ioreal_t* std_map)
{
   osivx_int scode = 0;

   osivx_int wind_xpos, wind_ypos;
   osivx_int wind_xsize, wind_ysize;
   osivx_int ind, ix, iy;
   osivx_real vec[2];  

   osivx_immw_t* map = (osivx_immw_t*)alg->udata[0];

   /* offset the first image */
   im0 -= p->disp_xoffset + p->disp_yoffset * p->image_xsize;

   /* find the window position */
   scode = osivx_grid_iter_get_info(iter,
      &wind_xpos, &wind_ypos,
      &wind_xsize, &wind_ysize,
      NULL, NULL);
   if (scode) goto finished;

   /* displace to window coordinates */
   im0 += wind_xpos + wind_ypos * p->image_xsize;
   im1 += wind_xpos + wind_ypos * p->image_xsize;

   /* calculate maps */
   osivx_alg1_getmap(p, im0, im1,
      wind_xsize, wind_ysize, map, &ind);

   /* calculate actual displacement */
   ix = ind % p->map_xsize;
   iy = ind / p->map_xsize;

   *(vec) = ix + p->disp_xoffset - p->disp_xmax;
   *(vec + 1) = iy + p->disp_yoffset - p->disp_ymax;

   /* perform interpolation */
   if (ix > 0 && ix < p->map_xsize - 1 &&
       iy > 0 && iy < p->map_ysize - 1) {
      osivx_interp_min_immw(p->interp_data,
        map + ind, -1, 1, -p->map_xsize,
        p->map_xsize, vec, vec + 1);
   }

   /* copy over found displacement */
   *(std_vec) = *(vec);
   *(std_vec + 1) = *(vec + 1);

   /* copy over correlation map */
   if (std_map != NULL) {
      osivx_alg1_map2std(p, std_map, map);
   }

finished:
   return scode;
}


/**********************************************************
 * The following function frees any memory that was
 * generated by an algorithm 1 instance.  It cannot
 * generate an error.
 **********************************************************/
static void
osivx_alg1_do_free(
   osivx_wind_alg_t* alg)
{
   if (alg->udata[0] != NULL) {
      osivx_free(alg->udata[0]);
   }
}
   

/**********************************************************
 * The following function assigns all of the hooks
 * that will be needed for running alg1. It cannot
 * generate an error.
 **********************************************************/
void
osivx_alg1_make(
   osivx_wind_alg_t* alg)
{
   alg->do_glbl_init = NULL;
   alg->do_init = osivx_alg1_do_init;
   alg->do_window = osivx_alg1_do_window;
   alg->do_free = osivx_alg1_do_free;
   alg->do_glbl_free = NULL;
}


