/* FILE: alg2.c 
 * 
 * Copyright (c) 2002 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"

osivx_int
osivx_alg2_checkargs(
  osivx_args_t * a)
{
   /* check for grid extending over first image */
   if (a->grid_xstart - a->disp_xoffset - a->disp_xmax - 1 < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart - a->disp_xoffset + a->disp_xmax +
        (a->grid_xsize - 1) * a->grid_xspacing +
        a->wind_xsize >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart - a->disp_yoffset - a->disp_ymax - 1 < 0) {
      return OSIVX_EYGRID;
   }
   if (a->grid_ystart - a->disp_yoffset + a->disp_ymax +
        (a->grid_ysize - 1) * a->grid_yspacing +
        a->wind_ysize >= a->image_ysize) {
     return OSIVX_EYGRID;
   }

   /* check for grid extending over second image */
   if (a->grid_xstart - 1 < 0) {
      return OSIVX_EXGRID;
   }
   if (a->grid_xstart + a->wind_xsize +
       (a->grid_xsize - 1) * a->grid_xspacing
        >= a->image_xsize) {
      return OSIVX_EXGRID;
   }

   if (a->grid_ystart - 1 < 0) {
      return OSIVX_EYGRID;
   }
   if (a->grid_ystart + a->wind_ysize +
       (a->grid_ysize - 1) * a->grid_yspacing
        >= a->image_ysize) {
      return OSIVX_EYGRID;
   }

   return 0;
}

static void
osivx_alg2_getint(
  osivx_opts_t * p,
  osivx_imcp_t * im0,
  osivx_imcp_t * im1,
  osivx_immf_t * sqd)
{
  osivx_int i, j;

  /* set first pixel */
  {
    osivx_imsp_t d = *im0 - *im1;
    *sqd = d * d;
  }

  /* integrate first row */
  for (i = 1; i < p->dom_xsize; i++)
    {
      osivx_imsp_t d = *(im0 + i) - *(im1 + i);
      *(sqd + i) = *(sqd + i - 1) + d * d;
    }

  /* integrate rest of map */
  for (i = 1; i < p->dom_ysize; i++)

    {
      osivx_imsp_t d;
      osivx_immf_t *sr = sqd + i * p->dom_xsize;
      osivx_imcp_t *i0 = im0 + i * p->image_xsize;
      osivx_imcp_t *i1 = im1 + i * p->image_xsize;

      d = *i0 - *i1;
      *sr = d * d;

      for (j = 1; j < p->dom_xsize; j++)
        {
          osivx_immf_t *sqdp = sr + j - 1;

          d = *(i0 + j) - *(i1 + j);
          *(sqdp + 1) = *sqdp + d * d;
          *sqdp += *(sqdp - p->dom_xsize);
        }
      *(sr + j - 1) += *(sr + j - 1 - p->dom_xsize);
    }
}

static void
osivx_alg2_getwind(
  osivx_opts_t * p,
  osivx_immf_t * sqd,
  osivx_immw_t * map)
{
  osivx_int i, j;
  osivx_int d1, d2, d3;

  /* calculate relative displacement */
  d1 = p->wind_xsize + p->wind_ysize * p->dom_xsize;
  d2 = p->wind_xsize;
  d3 = p->wind_ysize * p->dom_xsize;

  /* traverse array */
  for (i = 0; i < p->grid_ysize; i++)
    {
      osivx_int m = i * p->grid_yspacing * p->dom_xsize;
      for (j = 0; j < p->grid_xsize; j++)
        {
          osivx_int n = m + j * p->grid_xspacing;
          *map = *(sqd + n) + *(sqd + n + d1) -
             *(sqd + n + d2) - *(sqd + n + d3);
          map += p->map_len;
        }
    }
}

static void
osivx_alg2_map2std(
  osivx_opts_t * p,
  osivx_ioreal_t * dest,
  osivx_immw_t * src)
{
  osivx_int i;

  for (i = p->map_len - 1; i >= 0; i--)
    *(dest + i) = (osivx_ioreal_t) * (src + i);
}

static void
osivx_alg2_vec2std(
  osivx_opts_t * p,
  osivx_ioreal_t * dest,
  osivx_real * src)
{
#ifdef OSIVX_SWAP_REAL
  osivx_int i;

  for (i = 2 * p->grid_len - 1; i >= 0; i--)
    *(dest + i) = (osivx_ioreal_t) * (src + i);
#endif
}

osivx_int
osivx_alg2(
  osivx_opts_t * p)
{
  osivx_int i;
  osivx_int scode = 0;

  osivx_imcp_t *im0 = NULL;
  osivx_imcp_t *im1 = NULL;
  osivx_immf_t *sqd = NULL;
  osivx_immw_t *map = NULL;
  osivx_real *vec = NULL;

  osivx_ioreal_t *map_std = NULL;
  osivx_ioreal_t *vec_std = NULL;

  /* allocate memory */
  im0 = (osivx_imcp_t *) osivx_malloc(p->image_len * sizeof(osivx_imcp_t));
  im1 = (osivx_imcp_t *) osivx_malloc(p->image_len * sizeof(osivx_imcp_t));

  sqd = (osivx_immf_t *) osivx_malloc(p->dom_len * sizeof(osivx_immf_t));

  map = (osivx_immw_t *) osivx_malloc(p->grid_len * p->map_len * sizeof(osivx_immw_t));
  if (p->write_map)
    map_std = (osivx_ioreal_t *) osivx_malloc(p->map_len * sizeof(osivx_ioreal_t));

  vec = (osivx_real *) osivx_malloc(2 * p->grid_len * sizeof(osivx_real));
#ifdef OSIVX_SWAP_REAL
  vec_std = (osivx_ioreal_t *) osivx_malloc(2 * p->grid_len * sizeof(osivx_ioreal_t));
#else
  vec_std = (osivx_ioreal_t *) vec;
#endif

  if (!im0 || !im1 || !sqd || !map || !vec || !vec_std || (p->write_map && !map_std))
    {
      scode = OSIVX_EMEM;
      goto osivx_alg2_exit;
    }

  /* open body tag */
  if ((scode = osivx_fwrite_tag(p->vectors_file, OSIVX_TAG_BODY)))
    goto osivx_alg2_exit;

  /* run through images */
  for (i = p->movie_start; i <= p->movie_finish; i += p->movie_skip)
    {
      osivx_int j, k;
      osivx_imcp_t *im0_ptr;
      osivx_imcp_t *im1_ptr;
      osivx_immw_t *map_ptr = map;
      osivx_real *vec_ptr = vec;

      /* read the ith image */
      if ((scode = osivx_movie_read(p->movie_afile, i, im0)) ||
	  (scode = osivx_preproc_exec(p->preproc_adata, i, im0)))
	goto osivx_alg2_exit;

      /* read the ith image pair */
      if ((scode = osivx_movie_read(p->movie_bfile, i + p->movie_set, im1)) ||
          (scode = osivx_preproc_exec(p->preproc_bdata, i + p->movie_set, im1)))
        goto osivx_alg2_exit;

      /* open frame tag */
      if ((scode = osivx_fwrite_tag(p->vectors_file, OSIVX_TAG_FRAME)))
        goto osivx_alg2_exit;

      /* write frame number */
      if ((scode = osivx_fwrite_framenum(p->vectors_file, i)))
        goto osivx_alg2_exit;

      /* set the start locations */
      im0_ptr = im0 + (p->grid_xstart - p->disp_xoffset - 1) +
        (p->grid_ystart - p->disp_yoffset - 1) * p->image_xsize;
      im1_ptr = im1 + (p->grid_xstart - 1) + (p->grid_ystart - 1) * p->image_xsize;

      /* run through each displacement */
      for (j = p->disp_ymax; j >= -p->disp_ymax; j--)
        {
          osivx_int m = j * p->image_xsize;
          for (k = p->disp_xmax; k >= -p->disp_xmax; k--)
            {
              /* calculate the integral */
              osivx_alg2_getint(p, im0_ptr + m + k, im1_ptr, sqd);

              /* use the integral to calculate error */
              osivx_alg2_getwind(p, sqd, map_ptr++);
            }
        }

      /* reset map pointer */
      map_ptr = map;

      /* find the minimum error value */
      for (j = 0; j < p->grid_len; j++)
        {
          osivx_int ind;
          osivx_int ix, iy;

          /* find location of minima */
          ind = osivx_find_min_immw(map_ptr, p->map_len);
          ix = ind % p->map_xsize;
          iy = ind / p->map_xsize;

          /* calculate actual displacement */
          *(vec_ptr) = ix + p->disp_xoffset - p->disp_xmax;
          *(vec_ptr + 1) = iy + p->disp_yoffset - p->disp_ymax;

          /* perform interpolation */
          if (ix > 0 && ix < p->map_xsize - 1 &&
	      iy > 0 && iy < p->map_ysize - 1)
            {
              osivx_interp_min_immw(p->interp_data, map_ptr + ind, -1, 1,
                                   -p->map_xsize, p->map_xsize,
				   vec_ptr, vec_ptr + 1);
            }

          /* write error map to file */
          if (p->write_map)
            {
              osivx_alg2_map2std(p, map_std, map_ptr);

              if ((scode = osivx_fwrite_tag(p->vectors_file, OSIVX_TAG_MAPBOX)) ||
                  (scode = osivx_fwrite_mapnum(p->vectors_file, j)) ||
                  (scode = osivx_fwrite_map(p->vectors_file, p->map_len, map_std)) ||
                  (scode = osivx_fwrite_tag(p->vectors_file, OSIVX_GAT_MAPBOX)))
                goto osivx_alg2_exit;
            }

          vec_ptr += 2;
          map_ptr += p->map_len;
        }

      /* write vectors to file */
      osivx_alg2_vec2std(p, vec_std, vec);

      if ((scode = osivx_fwrite_vectors(p->vectors_file, 2 * p->grid_len, vec_std)))
        goto osivx_alg2_exit;

      /* close frame tag */
      if ((scode = osivx_fwrite_tag(p->vectors_file, OSIVX_GAT_FRAME)))
        goto osivx_alg2_exit;
      }

  /* close body tag */
  if ((scode = osivx_fwrite_tag(p->vectors_file, OSIVX_GAT_BODY)))
    goto osivx_alg2_exit;

osivx_alg2_exit:

  /* free memory */
  if (im0)
    osivx_free(im0);
  if (im1)
    osivx_free(im1);
  if (map)
    osivx_free(map);
  if (sqd)
    osivx_free(sqd);
  if (vec)
    osivx_free(vec);
#ifdef OSIVX_SWAP_REAL
  if (vec_std)
    osivx_free(vec_std);
#endif
  if (map_std)
    osivx_free(map_std);

  return scode;
}
