/* FILE: grid.c
 *
 * Copyright (c) 2006 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivx-devel.h"

#include <stdlib.h>


/**********************************************************
 * The following function generates a grid object of
 * the rectangular type.  It returns a NULL pointer
 * on error.
 **********************************************************/
osivx_grid_t*
osivx_grid_init_rect(
   osivx_int xstart,
   osivx_int ystart,
   osivx_int xspacing,
   osivx_int yspacing,
   osivx_int xsize,
   osivx_int ysize,
   osivx_int wind_xsize,
   osivx_int wind_ysize,
   osivx_int wind_xsuper,
   osivx_int wind_ysuper)
{
   osivx_grid_t* out;

   out = (osivx_grid_t*)malloc(sizeof(osivx_grid_t));
   if (out == NULL) {
      return NULL;
   }

   out->type = OSIVX_GRID_RECT;
   out->u.rect.xstart = xstart;
   out->u.rect.ystart = ystart;
   out->u.rect.xspacing = xspacing;
   out->u.rect.yspacing = yspacing;
   out->u.rect.xsize = xsize;
   out->u.rect.ysize = ysize;
   out->u.rect.wind_xsize = wind_xsize;
   out->u.rect.wind_ysize = wind_ysize;
   out->u.rect.wind_xsuper = wind_xsuper;
   out->u.rect.wind_ysuper = wind_ysuper;

   return out; 
}


/**********************************************************
 * The following function frees a grid object.  It
 * returns non-zero if an errors is encountered.
 **********************************************************/
osivx_int
osivx_grid_free(
   osivx_grid_t* grid)
{
   osivx_free(grid);

   return 0;
}


/**********************************************************
 * The following function resets the passed iterator
 * to point to the beginning of the grid list.  This 
 * function cannot generate errors.
 **********************************************************/
void
osivx_grid_begin(
   const osivx_grid_t* grid,
   osivx_grid_iter_t* iter)
{
   iter->grid_ptr = grid;

   switch (grid->type) {
   case OSIVX_GRID_RECT:
      iter->u.rect.pos = 0;
      break;
   }
}


/**********************************************************
 * The following function returns non-zero if the iter 
 * is at the end of the grid list.
 **********************************************************/
osivx_int
osivx_grid_iter_at_end(
   const osivx_grid_iter_t* iter)
{
   const osivx_grid_t* grid = iter->grid_ptr;

   switch (grid->type) {
   case OSIVX_GRID_RECT:
      if (iter->u.rect.pos >=
         grid->u.rect.xsize *
         grid->u.rect.ysize) {
         return 1;
      } else {
         return 0;
      }
      break;
   }

   return 0;
}


/**********************************************************
 * The following function acquires the coordinates for
 * the current position of the grid iterator.  It sets
 * *xpos to the position of the leftmost pixel of the
 * current window, *ypos to the position of the rightmost
 * pixel of the current window, *xsize to the x window
 * size, *ysize to the y window size, *xsuper to the 
 * x super window size, and *ysuper to the y super window
 * size.  It returns a non-zero error code if the iterator
 * is at the end of the list;
 **********************************************************/
osivx_int
osivx_grid_iter_get_info(
   const osivx_grid_iter_t* iter,
   osivx_int* xpos, osivx_int* ypos, 
   osivx_int* xsize, osivx_int* ysize,
   osivx_int* xsuper, osivx_int* ysuper)
{
   const osivx_grid_t* grid = iter->grid_ptr;

   switch (grid->type) {
   case OSIVX_GRID_RECT:
      if (iter->u.rect.pos >=
         grid->u.rect.xsize *
         grid->u.rect.ysize) {
         return 1;
      }
      *xpos = (iter->u.rect.pos %
        grid->u.rect.xsize) *
        grid->u.rect.xspacing + 
        grid->u.rect.xstart;
      *ypos = (iter->u.rect.pos /
        grid->u.rect.xsize) *
        grid->u.rect.yspacing + 
        grid->u.rect.ystart;

      *xsize = grid->u.rect.wind_xsize;
      *ysize = grid->u.rect.wind_ysize;

      if (xsuper != NULL) {
         *xsuper = grid->u.rect.wind_xsuper;
      }
      if (ysuper != NULL) {
         *ysuper = grid->u.rect.wind_ysuper;
      }

      break;
   }

   return 0;
}


/**********************************************************
 * The following function repositions the grid iterator
 * so that it points to the next window in the grid. It
 * returns a non-zero number if the end of the list is
 * encountered.
 **********************************************************/
osivx_int
osivx_grid_iter_next(
   osivx_grid_iter_t* iter)
{
   const osivx_grid_t* grid = iter->grid_ptr;

   switch (grid->type) {
   case OSIVX_GRID_RECT: 
      if (iter->u.rect.pos >=
         grid->u.rect.xsize *
         grid->u.rect.ysize) {
         return 1;
      } else {
         iter->u.rect.pos += 1;
      }
      break;
   }

   return 0;
}


/**********************************************************
 * The following function returns the total number of
 * windows in the given grid.  It returns a negative
 * number if an error is encountered.
 **********************************************************/
osivx_int
osivx_grid_get_len(
   const osivx_grid_t* grid)
{
   switch (grid->type) {
   case OSIVX_GRID_RECT:
      return grid->u.rect.xsize * grid->u.rect.ysize;
   default:
      return -1;
   }
}


/**********************************************************
 * The following function assigns the maximum sizes of
 * the x and y windows.  It returns zero no success and
 * a negative number if an error is encountered.
 **********************************************************/
osivx_int
osivx_grid_get_max_wind_sizes(
   const osivx_grid_t* grid,
   osivx_int* wind_xsize,
   osivx_int* wind_ysize,
   osivx_int* wind_xsuper,
   osivx_int* wind_ysuper)
{
   switch (grid->type) {
   case OSIVX_GRID_RECT:
      *wind_xsize = grid->u.rect.wind_xsize;
      *wind_ysize = grid->u.rect.wind_ysize;

      if (wind_xsuper != NULL) {
         *wind_xsuper = grid->u.rect.wind_xsuper;
      }
      if (wind_ysuper != NULL) {
         *wind_ysuper = grid->u.rect.wind_ysuper;
      }
      break;   
   default:
      return -1;
   }

   return 0;
}


/**********************************************************
 * The following function assigns the mode of the sizes 
 * of the x and y windows.  It returns zero no success and
 * a negative number if an error is encountered.
 **********************************************************/
osivx_int
osivx_grid_get_mode_wind_sizes(
   const osivx_grid_t* grid,
   osivx_int* wind_xsize,
   osivx_int* wind_ysize,
   osivx_int* wind_xsuper,
   osivx_int* wind_ysuper)
{
   switch (grid->type) {
   case OSIVX_GRID_RECT:
      *wind_xsize = grid->u.rect.wind_xsize;
      *wind_ysize = grid->u.rect.wind_ysize;

      if (wind_xsuper != NULL) {
         *wind_xsuper = grid->u.rect.wind_xsuper;
      }
      if (wind_ysuper != NULL) {
         *wind_ysuper = grid->u.rect.wind_ysuper;
      }
      break;   
   default:
      return -1;
   }

   return 0;
}


/**********************************************************
 * The following function finds the range for a grid,
 * the bounding box for all of the windows.  It returns
 * a negative number if an error is encountered.
 **********************************************************/
osivx_int
osivx_grid_get_range(
   const osivx_grid_t* grid,
   osivx_int* xstart,
   osivx_int* ystart,
   osivx_int* xsize,
   osivx_int* ysize)
{
   switch (grid->type) {
   case OSIVX_GRID_RECT:
      *xstart = grid->u.rect.xstart;
      *ystart = grid->u.rect.ystart;
      *xsize = (grid->u.rect.xsize - 1) *
        grid->u.rect.xspacing +
        grid->u.rect.wind_xsize;
      *ysize = (grid->u.rect.ysize - 1) *
        grid->u.rect.yspacing +
        grid->u.rect.wind_ysize;
      return 0;
   default:
      return -1;
   }
}


