/* FILE: load.c
 *
 * Copyright (c) 2002-2006 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivx-devel.h"


/**********************************************************
 * The following function performs a raw read from the
 * osivx_file_t.  It does any byte or bit swapping that may
 * be necessary to convert to the native platform.  It
 * returns 0 on success and an error code on failure.
 **********************************************************/
static osivx_int
osivx_fread(
   osivx_file_t* file,
   size_t size,
   size_t nitems,
   void *ptr)
{

   if (fread(ptr, size, nitems, file->stream) != nitems) {
      return OSIVX_EFILEREAD;
   }

   if (file->swap_bytes) {
      switch (size) {
      case 1:
         break;
      case 2: {
         const uint16_t mask = 0xff;

         uint16_t val0, val1;
         uint16_t *bptr = (uint16_t *) ptr;
         size_t count = nitems;

         while (count--) {
            val0 = *bptr;
            val1 = (val0 & (mask << 0)) << 8;
            val1 |= (val0 & (mask << 8)) >> 8;
            *bptr++ = val1;
         }
      } break;
      case 4: {
         register size_t count = nitems;
         register uint32_t val0, val1;
         register uint32_t *bptr = (uint32_t *) ptr;

         while (count--) {
            const uint32_t mask = 0xff;

            val0 = *bptr;
            val1 = (val0 & (mask << 0)) << 24;
            val1 |= (val0 & (mask << 8)) << 8;
            val1 |= (val0 & (mask << 16)) >> 8;
            val1 |= (val0 & (mask << 24)) >> 24;
            *bptr++ = val1;
         }
      } break;
      case 8: {
         register size_t count = nitems;
         register uint64_t val0, val1;
         register uint64_t *bptr = (uint64_t *) ptr;

         while (count--) {
            const uint64_t mask = 0xff;

            val0 = *bptr;
            val1 = (val0 & (mask << 0)) << 56;
            val1 |= (val0 & (mask << 8)) << 40;
            val1 |= (val0 & (mask << 16)) << 24;
            val1 |= (val0 & (mask << 24)) << 8;
            val1 |= (val0 & (mask << 32)) >> 8;
            val1 |= (val0 & (mask << 40)) >> 24;
            val1 |= (val0 & (mask << 48)) >> 40;
            val1 |= (val0 & (mask << 56)) >> 56;
            *bptr++ = val1;
         }
      } break;
      default: {
         register char *bptr = (char *) ptr;
         register size_t count = nitems;

         while (count--) {
            register size_t i;
            for (i = 0; i < size / 2; i++) {
               char val0 = *(bptr + i);
               *(bptr + i) = *(bptr + size - i - 1);
               *(bptr + size - i - 1) = val0;
            }
            bptr += size;
         }
      } break;
      }
   }

   if (file->swap_bits) {
      register char c0, c1;
      register char *bptr = (char *) ptr;
      register size_t count = size * nitems;

      while (count--) {
         c0 = *bptr;
         c1 = (c0 & 0x01) << 7;
         c1 |= (c0 & 0x02) << 5;
         c1 |= (c0 & 0x04) << 3;
         c1 |= (c0 & 0x08) << 1;
         c1 |= (c0 & 0x10) >> 7;
         c1 |= (c0 & 0x20) >> 5;
         c1 |= (c0 & 0x40) >> 3;
         c1 |= (c0 & 0x80) >> 1;
         *bptr++ = c1;
      }
   }

   return 0;
}


/**********************************************************
 * The following function reads the file IO version from
 * the given file.  It does not require seek capabilities.
 * It returns 0 on success and an error code on failure.
 **********************************************************/
osivx_int
osivx_fread_ver(
   osivx_file_t* fp,
   int *ver)
{
   ver[0] = fp->ver_first;
   ver[1] = fp->ver_second;
   ver[2] = fp->ver_third;
   return 0;
}


/**********************************************************
 * The following function reads a tag from the osivx_file_t.
 * It does not require seek capabilities.  It returns 0
 * on success and an error code on failure.
 **********************************************************/
osivx_int
osivx_fread_tag(
   osivx_file_t* file,
   osivx_iotag_t * tag)
{
   return osivx_fread(file, sizeof(osivx_iotag_t), 1, tag);
}


/**********************************************************
 * The following function reads the contents of an integer
 * argument box from the osivx_file_t.  It does not require
 * seek capabilities.  It returns 0 on success and an
 * error code on failure.
 **********************************************************/
osivx_int
osivx_fread_argint(
   osivx_file_t* file,
   osivx_ioint_t* key,
   osivx_ioint_t * val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;
   
   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, key);
   if (scode) return scode;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, val);
   if (scode) return scode;

   if (blen != 2 * sizeof(osivx_ioint_t)) {
      return OSIVX_EFILEREAD;
   }

   return 0;
}


/**********************************************************
 * The following function reads the contents of a string
 * argument box from the osivx_file_t.  It does not require
 * seek capabilities.  It returns 0 on success and an
 * error code on failure.
 **********************************************************/
osivx_int
osivx_fread_argstr(osivx_file_t * file, osivx_ioint_t * key, char **val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   *val = NULL;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) goto finished;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, key);
   if (scode) goto finished;

   if ((blen -= sizeof(osivx_ioint_t)) < 1) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   *val = (char *)osivx_malloc(blen * sizeof(char));
   if (*val == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }
   
   scode = osivx_fread(file, sizeof(char), blen, *val);
   if (scode) goto finished;

   if (*(*val + blen - 1) != '\0') {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

finished:

   if (scode) {
      if (*val) {
         osivx_free(*val);
      }
   }

   return scode;
}


/**********************************************************
 * The following function reads the contents of a frame
 * number box from the osivx_file_t.  It does not require
 * seek capabilities.  It returns 0 on success and an
 * error code on failure.
 **********************************************************/
osivx_int
osivx_fread_framenum(
   osivx_file_t* file,
   osivx_ioint_t* val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, val);
   if (scode) return scode;

   if (blen != sizeof(osivx_ioint_t)) {
      return OSIVX_EFILEREAD;
   }

   return scode;
}


/**********************************************************
 * The following function reads the contents of a vectors
 * box from the osivx_file_t.  It does not require seek
 * capabilities.  It returns 0 on success and an error
 * code on failure.
 **********************************************************/
osivx_int
osivx_fread_vectors(
   osivx_file_t* file,
   osivx_ioint_t* len,
   osivx_ioreal_t** vecs)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   *vecs = NULL;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) goto finished;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, len);
   if (scode) goto finished;

   if (blen != sizeof(osivx_ioint_t) +
       (*len)*sizeof(osivx_ioreal_t)) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   *vecs = (osivx_ioreal_t*)osivx_malloc(
     (*len) * sizeof(osivx_ioreal_t));
   if (*vecs == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   scode = osivx_fread(file, sizeof(osivx_ioreal_t), *len, *vecs);
   if (scode) goto finished;


finished:
   if (scode) {
      if (*vecs) {
         osivx_free(*vecs);
      }
   }

   return scode;
}


/**********************************************************
 * The following function reads the contents of a mapnum
 * box from the osivx_file_t.  It does not require seek
 * capabilities.  It returns 0 on success and an error
 * code on failure.
 **********************************************************/
osivx_int
osivx_fread_mapnum(
   osivx_file_t* file,
   osivx_ioint_t * val)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) return scode;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, val);
   if (scode) return scode;

   if (blen != sizeof(osivx_ioint_t)) {
      return OSIVX_EFILEREAD;
   }

   return 0;
}


/**********************************************************
 * The following function reads the contents of a map
 * box from the osivx_file_t.  It does not require seek
 * capabilities.  It returns 0 on success and an error
 * code on failure.
 **********************************************************/
osivx_int
osivx_fread_map(
   osivx_file_t* file,
   osivx_ioint_t* len,
   osivx_ioreal_t** map)
{
   osivx_ioint_t blen;
   osivx_int scode = 0;

   *map = NULL;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
   if (scode) goto finished;

   scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, len);
   if (scode) goto finished;

   if (blen != sizeof(osivx_ioint_t) +
        (*len)*sizeof(osivx_ioreal_t)) {
      scode = OSIVX_EFILEREAD;
      goto finished;
   }

   *map = (osivx_ioreal_t*)osivx_malloc(
      (*len) * sizeof(osivx_ioreal_t));
   if (*map == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   scode = osivx_fread(file, sizeof(osivx_ioreal_t), *len, *map);
   if (scode) goto finished;

finished:
   if (scode) {
      if (*map) {
         osivx_free(*map);
      }
   }

   return scode;
}


/**********************************************************
 * The following function returns zero number if the tags
 * are equivalent and a non-zero number otherwise.
 **********************************************************/
osivx_int
osivx_tagcmp(osivx_iotag_t a, osivx_iotag_t b)
{
   return (a ^ b);
}


/**********************************************************
 * The following function scans the file forward to the
 * given tag.  It requires relative forwarding seek
 * capabilities from the osivx_file_t.  It returns 0 on
 * success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fscan_totag(
   osivx_file_t* file,
   osivx_iotag_t tag)
{
   osivx_iotag_t top;
   osivx_int scode = 0;

   while (OSIVX_TRUE) {
      scode = osivx_fread(file, sizeof(osivx_iotag_t), 1, &top);
      if (scode) return scode;

      if (top == tag) {
         return 0;
      }

      if (top & OSIVX_TAG_ISCLOSE) {
         return OSIVX_EFILEREAD;
      } else if (top & OSIVX_TAG_ISBASE) {
         osivx_ioint_t blen;
         osivx_iotag_t bottom;

         scode = osivx_fread(file, sizeof(osivx_ioint_t), 1, &blen);
         if (scode) return scode;

         scode = osivx_fseek(file, blen, SEEK_CUR);
         if (scode) return scode;

         scode = osivx_fread(file, sizeof(osivx_iotag_t), 1, &bottom);
         if (scode) return scode;

         if (osivx_tagcmp(bottom, ~top)) {
            return OSIVX_EFILEREAD;
         }
      } else {
         scode = osivx_fscan_totag(file, ~top);
         if (scode) return scode;
      }
   }
}


/**********************************************************
 * The following function scans the file forward to the
 * given frame.  It assumes that the file is positioned
 * within the body element.  It requires relative
 * forwarding seek capabilities from the osivx_file_t.  It
 * returns 0 on success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fscan_toframe(
   osivx_file_t* file,
   osivx_int num)
{
   long pos;
   osivx_ioint_t current;
   osivx_iotag_t closer;
   osivx_int scode = 0;

   scode = osivx_fscan_totag(file, OSIVX_TAG_FRAME);
   if (scode) return scode;

   scode = osivx_ftell(file, &pos);
   if (scode) return scode;

   scode = osivx_fscan_totag(file, OSIVX_TAG_FRAMENUM);
   if (scode) return scode;

   scode = osivx_fread_framenum(file, &current);
   if (scode) return scode;

   scode = osivx_fread_tag(file, &closer);
   if (scode) return scode;

   if (osivx_tagcmp(closer, OSIVX_GAT_FRAMENUM)) {
      return OSIVX_EFILEREAD;
   }

   while (current != num) {
      scode = osivx_fscan_totag(file, OSIVX_GAT_FRAME);
      if (scode) return scode;

      scode = osivx_fscan_totag(file, OSIVX_TAG_FRAME);
      if (scode) return scode;

      scode = osivx_ftell(file, &pos);
      if (scode) return scode;

      scode = osivx_fscan_totag(file, OSIVX_TAG_FRAMENUM);
      if (scode) return scode;

      scode = osivx_fread_framenum(file, &current);
      if (scode) return scode;

      scode = osivx_fread_tag(file, &closer);
      if (scode) return scode;

      if (osivx_tagcmp(closer, OSIVX_GAT_FRAMENUM)) {
         return OSIVX_EFILEREAD;
      }
   }

   return osivx_fseek(file, pos, SEEK_SET);
}


/**********************************************************
 * The following function scans the file forward to the
 * given map.  It assumes that the file is positioned
 * within a frame element.  It requires relative forwarding
 * seek capabilities from the osivx_file_t.  It returns 0 on
 * success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fscan_tomap(
   osivx_file_t* file,
   osivx_int num)
{
   long pos;
   osivx_ioint_t current;
   osivx_iotag_t closer;
   osivx_int scode = 0;

   scode = osivx_fscan_totag(file, OSIVX_TAG_MAPBOX);
   if (scode) return scode;

   scode = osivx_ftell(file, &pos);
   if (scode) return scode;

   scode = osivx_fscan_totag(file, OSIVX_TAG_MAPNUM);
   if (scode) return scode;

   scode = osivx_fread_mapnum(file, &current);
   if (scode) return scode;

   scode = osivx_fread_tag(file, &closer);
   if (scode) return scode;

   if (osivx_tagcmp(closer, OSIVX_GAT_MAPNUM)) {
      return OSIVX_EFILEREAD;
   }

   while (current != num) {
      scode = osivx_fscan_totag(file, OSIVX_GAT_MAPBOX);
      if (scode) return scode;

      scode = osivx_fscan_totag(file, OSIVX_TAG_MAPBOX);
      if (scode) return scode;

      scode = osivx_ftell(file, &pos);
      if (scode) return scode;

      scode = osivx_fscan_totag(file, OSIVX_TAG_MAPNUM);
      if (scode) return scode;

      scode = osivx_fread_mapnum(file, &current);
      if (scode) return scode;

      scode = osivx_fread_tag(file, &closer);
      if (scode) return scode;

      if (osivx_tagcmp(closer, OSIVX_GAT_MAPNUM)) {
         return OSIVX_EFILEREAD;
      }
   }

   return osivx_fseek(file, pos, SEEK_SET);
}


/**********************************************************
 * The following function grabs a particular integer
 * argument from the file.  It assumes that the file is
 * positioned at the beginning of the header, and leaves
 * the file positioned in the original spot.  It requires
 * absolute fseek capabilities from the osivx_file_t.  It
 * returns 0 on success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fgrab_argint(
   osivx_file_t* file,
   osivx_ioint_t key,
   osivx_ioint_t* val)
{
   long pos;
   osivx_ioint_t current;
   osivx_iotag_t closer;
   osivx_int scode = 0;

   scode = osivx_ftell(file, &pos);
   if (scode) return scode;

   do {
      scode = osivx_fscan_totag(file, OSIVX_TAG_ARGINT);
      if (scode) return scode;

      scode = osivx_fread_argint(file, &current, val);
      if (scode) return scode;

      scode = osivx_fread_tag(file, &closer);
      if (scode) return scode;

      if (osivx_tagcmp(closer, OSIVX_GAT_ARGINT)) {
         return OSIVX_EFILEREAD;
      }
   } while (current != key);

   return osivx_fseek(file, pos, SEEK_SET);
}


/**********************************************************
 * The following function grabs a particular string
 * argument from the file.  It assumes that the file is
 * positioned at the beginning of the header, and leaves
 * the file positioned in the original spot.  It requires
 * absolute fseek capabilities from the osivx_file_t.  It
 * returns 0 on success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fgrab_argstr(
   osivx_file_t* file,
   osivx_ioint_t key,
   char** val)
{
   long pos;
   osivx_ioint_t current;
   osivx_iotag_t closer;
   osivx_int scode = 0;

   scode = osivx_ftell(file, &pos);
   if (scode) return scode;

   do {
      scode = osivx_fscan_totag(file, OSIVX_TAG_ARGSTR);
      if (scode) return scode;

      scode = osivx_fread_argstr(file, &current, val);
      if (scode) return scode;
      
      scode = osivx_fread_tag(file, &closer);
      if (scode) return scode;

      if (osivx_tagcmp(closer, OSIVX_GAT_ARGSTR)) {
         return OSIVX_EFILEREAD;
      }
   } while (current != key);

   return osivx_fseek(file, pos, SEEK_SET);
}


/**********************************************************
 * The following function grabs a particular vectors
 * box from the file.  It assumes that the file is
 * positioned at the beginning of the body, and leaves
 * the file positioned in the original spot.  It requires
 * absolute fseek capabilities from the osivx_file_t.  It
 * returns 0 on success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fgrab_vectors(
   osivx_file_t* file,
   osivx_int frame,
   osivx_ioint_t* vectors_len,
   osivx_ioreal_t** vectors)
{
   long pos;
   osivx_iotag_t closer;
   osivx_int scode = 0;

   scode = osivx_ftell(file, &pos);
   if (scode) return scode;

   scode = osivx_fscan_toframe(file, frame);
   if (scode) return scode;

   scode = osivx_fscan_totag(file, OSIVX_TAG_VECTORS);
   if (scode) return scode;

   scode = osivx_fread_vectors(file, vectors_len, vectors);
   if (scode) return scode;

   scode = osivx_fread_tag(file, &closer);
   if (scode) return scode;

   if (osivx_tagcmp(closer, OSIVX_GAT_VECTORS)) {
      return OSIVX_EFILEREAD;
   }

   return osivx_fseek(file, pos, SEEK_SET);
}


/**********************************************************
 * The following function grabs a particular map
 * box from the file.  It assumes that the file is
 * positioned at the beginning of the body, and leaves
 * the file positioned in the original spot.  It requires
 * absolute fseek capabilities from the osivx_file_t.  It
 * returns 0 on success and an error code on failure. 
 **********************************************************/
osivx_int
osivx_fgrab_map(
   osivx_file_t* file,
   osivx_int frame,
   osivx_int window,
   osivx_ioint_t* map_len,
   osivx_ioreal_t** map)
{
   long pos;
   osivx_iotag_t closer;
   osivx_int scode = 0;

   scode = osivx_ftell(file, &pos);
   if (scode) return scode;

   scode = osivx_fscan_toframe(file, frame);
   if (scode) return scode;

   scode = osivx_fscan_tomap(file, window);
   if (scode) return scode;

   scode = osivx_fscan_totag(file, OSIVX_TAG_MAP);
   if (scode) return scode;

   scode = osivx_fread_map(file, map_len, map);
   if (scode) return scode;

   scode = osivx_fread_tag(file, &closer);
   if (scode) return scode;

   if (osivx_tagcmp(closer, OSIVX_GAT_MAP)) {
      return OSIVX_EFILEREAD;
   }

   return osivx_fseek(file, pos, SEEK_SET);
}

