/* FILE: movie.c
 *
 * Copyright (c) 2002 James A. Strother
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "osivx-devel.h"

#include <string.h>
#include <tiffio.h>

/* make movie types */
enum
  {
    OSIV_MOVIE_BAD,
    OSIV_MOVIE_STREAM, 
    OSIV_MOVIE_FILE, 
    OSIV_MOVIE_TIFF_MERGE, 
    OSIV_MOVIE_TIFF_SPLIT 
  };


static osivx_int
osivx_movie_gettype(
  const char *str)
{
  char *suf;

  if (!strcmp(str, "-"))
    return OSIV_MOVIE_STREAM;

  if (strrchr(str, '%') != strchr(str, '%'))
    return OSIV_MOVIE_BAD;

  suf = strrchr(str, '.');
  if (!suf++)
    return OSIV_MOVIE_BAD;

  if (!strcmp(suf, "dat"))
    return OSIV_MOVIE_FILE;

  if (!strcmp(suf, "tif") || !strcmp(suf, "tiff"))
    {
      if (strchr(str, '%'))
        return OSIV_MOVIE_TIFF_SPLIT;
      else
        return OSIV_MOVIE_TIFF_MERGE;
    }

  return OSIV_MOVIE_BAD;
}

static osivx_int
osivx_movie_openstream(
  osivx_movie_t * movie,
  const char *fname,
  osivx_int xsize,
  osivx_int ysize,
  osivx_int firstf,
  osivx_int skipf,
  osivx_int lastf)
{
  movie->file = (void *) stdin;
  return 0;
}

static osivx_int
osivx_movie_openfile(
  osivx_movie_t * movie,
  const char *fname,
  osivx_int xsize,
  osivx_int ysize,
  osivx_int firstf,
  osivx_int skipf,
  osivx_int lastf)
{

  movie->file = (void *) fopen(fname, "rb");
  if (movie->file != NULL) {
    OSIVX_ERR_SETSTR(movie->err_info, fname);
    return OSIVX_EMOVXFILE;
  }

  return 0;
}

static osivx_int
osivx_movie_opentiffmerge(
  osivx_movie_t * movie,
  const char *fname,
  osivx_int xsize,
  osivx_int ysize,
  osivx_int firstf,
  osivx_int skipf,
  osivx_int lastf)
{
  movie->file = (void *) TIFFOpen(fname, "r");
  if (!movie->file) {
    OSIVX_ERR_SETSTR(movie->err_info, fname);
    return OSIVX_EMOVXFILE;
  }

  return 0;
}

static osivx_int
osivx_movie_opentiffsplit(
  osivx_movie_t * movie,
  const char *fname,
  osivx_int xsize,
  osivx_int ysize,
  osivx_int firstf,
  osivx_int skipf,
  osivx_int lastf)
{
  movie->file = (void *) osivx_malloc(strlen(fname) + 1);
  if (!movie->file)
    return OSIVX_EMEM;
  strcpy((char *) movie->file, fname);
  return 0;
}

osivx_movie_t *
osivx_movie_open(
  const char *fname,
  osivx_int xsize,
  osivx_int ysize,
  osivx_int firstf,
  osivx_int skipf,
  osivx_int lastf,
  osivx_err* err_info)
{
  osivx_movie_t *movie = NULL;
  osivx_int scode = 0;

  /* allocate movie */
  movie = (osivx_movie_t *) osivx_malloc(sizeof(osivx_movie_t));
  if (!movie)
    {
      scode = OSIVX_EMEM;
      goto osivx_movie_open_exit;
    }

  /* grab err info */
  movie->err_info = err_info;

  /* set the filename.
   * NOTE: this must be placed here for proper
   * error-handling to occur
   */
  movie->name = (char*)osivx_malloc(strlen(fname)+1);
  if (movie->name == NULL) {
     scode = OSIVX_EMEM;
     goto osivx_movie_open_exit;
  }
  strcpy(movie->name, fname);


  /* grab type */
  movie->type = osivx_movie_gettype(fname);
  if (movie->type == OSIV_MOVIE_BAD) {
     scode = OSIVX_EMOVTYPE;
     goto osivx_movie_open_exit;
  }

  /* set sizes */
  movie->xsize = xsize;
  movie->ysize = ysize;
  movie->imlen = xsize * ysize;
  movie->firstf = firstf;
  movie->skipf = skipf;
  movie->lastf = lastf;

  /* open necessary files */
  switch (movie->type)
    {
    case OSIV_MOVIE_STREAM:
      scode = osivx_movie_openstream(movie, fname,
         xsize, ysize, firstf, skipf, lastf);
      break;
    case OSIV_MOVIE_FILE:
      scode = osivx_movie_openfile(movie, fname,
         xsize, ysize, firstf, skipf, lastf);
      break;
    case OSIV_MOVIE_TIFF_MERGE:
      scode = osivx_movie_opentiffmerge(movie, fname,
         xsize, ysize, firstf, skipf, lastf);
      break;
    case OSIV_MOVIE_TIFF_SPLIT:
      scode = osivx_movie_opentiffsplit(movie, fname,
         xsize, ysize, firstf, skipf, lastf);
      break;
    default:
      scode = OSIVX_EMOVTYPE;
      break;
    }

osivx_movie_open_exit:
  if (scode)
    {
      if (movie->name)
        {
          osivx_free(movie->name);
        }

      if (movie)
	{
          osivx_free(movie);
          movie = NULL;
        }
    }

  return movie;
}


/*********************************************************
 * This function reads an image of the given size from the
 * given TIFF file.  It returns zero on success and an
 * error code on failure.  
 *********************************************************/
static osivx_int
osivx_movie_readtiff(
  TIFF * tif,
  osivx_int xsize,
  osivx_int ysize,
  osivx_imcp_t * image)
{
  tsize_t i;
  uint16_t nbits, nsamples;
  uint32_t length, width;

  TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &nbits);
  if (nbits != 8 * sizeof(osivx_imcp_t))
    return OSIVX_EIMAGEINT;

  TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &nsamples);
  if (nsamples != 1) 
    return OSIVX_EMOVTIFF;

  TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
  if (width != xsize)
    return OSIVX_EMOVXSIZE;

  TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &length);
  if (length != ysize)
    return OSIVX_EMOVYSIZE;

  if (TIFFIsTiled(tif))
    return OSIVX_EMOVTIFF;

  for (i = 0; i < ysize; i++)
    {
      if (TIFFReadScanline(tif, image, i, 0) < 0)
        return OSIVX_EMOVTIFF;
      image += xsize;
    }

  return 0;
}

static osivx_int
osivx_movie_readstream(
  osivx_movie_t * movie,
  osivx_int frm,
  osivx_imcp_t * im)
{
  FILE *fp = movie->file;

  if (fread(im, sizeof(osivx_imcp_t), movie->imlen, fp) != movie->imlen) {
    OSIVX_ERR_SETSTR(movie->err_info, movie->name);
    return OSIVX_EMOVRAW;
  }

  return 0;
}

static osivx_int
osivx_movie_readfile(
  osivx_movie_t * movie,
  osivx_int frm,
  osivx_imcp_t * im)
{
  FILE *fp = movie->file;

  if (fseek(fp, frm * movie->imlen * sizeof(osivx_imcp_t), SEEK_SET) ||
      fread(im, sizeof(osivx_imcp_t), movie->imlen, fp) != movie->imlen) {
    OSIVX_ERR_SETSTR(movie->err_info, movie->name);
    return OSIVX_EMOVRAW;
  }

  return 0;
}

static osivx_int
osivx_movie_readtiffmerge(
  osivx_movie_t * movie,
  osivx_int frm,
  osivx_imcp_t * im)
{
  TIFF *fp = movie->file;
  osivx_int scode;

  if (!TIFFSetDirectory(fp, frm)) {
    scode = OSIVX_EMOVDIR;
    goto osivx_movie_readtiffmerge_exit;
  }

  scode = osivx_movie_readtiff(fp, movie->xsize, movie->ysize, im);
  if (scode != 0) {
    goto osivx_movie_readtiffmerge_exit;
  }

osivx_movie_readtiffmerge_exit:
  if (scode != 0) {
    OSIVX_ERR_SETSTR(movie->err_info, movie->name);
  }

  return scode;
}

static osivx_int
osivx_movie_readtiffsplit(
  osivx_movie_t * movie,
  osivx_int frm,
  osivx_imcp_t * im)
{
  size_t len;
  TIFF *fp = NULL;
  char *str = NULL;
  osivx_int scode = 0;

  len = strlen(movie->file) + 8 * sizeof(osivx_int) + 1;
  str = (char *) osivx_malloc(len * sizeof(char));
  if (!str)
    {
      scode = OSIVX_EMEM;
      goto osivx_movie_readtiffsplit_exit;
    }

  sprintf(str, (char *) movie->file, frm);

  if (!(fp = TIFFOpen(str, "r")))
    {
      scode = OSIVX_EMOVTIFF;
      goto osivx_movie_readtiffsplit_exit;
    }

  scode = osivx_movie_readtiff(fp, movie->xsize, movie->ysize, im);
  if (scode != 0) {
     goto osivx_movie_readtiffsplit_exit;
  }

osivx_movie_readtiffsplit_exit:
  if (scode != 0) {
     if (str != NULL) {
        OSIVX_ERR_SETSTR(movie->err_info, str);
     } else {
        OSIVX_ERR_SETSTR(movie->err_info, "");
     }
  }

  if (str)
    osivx_free(str);
  if (fp)
    TIFFClose(fp);

  return scode;
}

osivx_int
osivx_movie_read(
  osivx_movie_t * movie,
  osivx_int frm,
  osivx_imcp_t * im)
{
  switch (movie->type)
    {
    case OSIV_MOVIE_STREAM:
      return osivx_movie_readstream(movie, frm, im);
    case OSIV_MOVIE_FILE:
      return osivx_movie_readfile(movie, frm, im);
    case OSIV_MOVIE_TIFF_MERGE:
      return osivx_movie_readtiffmerge(movie, frm, im);
    case OSIV_MOVIE_TIFF_SPLIT:
      return osivx_movie_readtiffsplit(movie, frm, im);
    default:
      return OSIVX_EMOVTYPE;
    }
}

static osivx_int
osivx_movie_closestream(
  osivx_movie_t * movie)
{
  if (fflush((FILE *) movie->file) == EOF) {
    OSIVX_ERR_SETSTR(movie->err_info, movie->name);
    return OSIVX_EMOVRAW;
  }

  osivx_free(movie->name);
  osivx_free(movie);
  return 0;
}

static osivx_int
osivx_movie_closefile(
  osivx_movie_t * movie)
{
  if (fclose((FILE *) movie->file) == EOF) {
    OSIVX_ERR_SETSTR(movie->err_info, movie->name);
    return OSIVX_EMOVRAW;
  }

  osivx_free(movie->name);
  osivx_free(movie);
  return 0;
}

static osivx_int
osivx_movie_closetiffmerge(
  osivx_movie_t * movie)
{
  TIFFClose((TIFF *) movie->file);
  osivx_free(movie->name);
  osivx_free(movie);
  return 0;
}

static osivx_int
osivx_movie_closetiffsplit(
  osivx_movie_t * movie)
{
  osivx_free(movie->name);
  osivx_free(movie->file);
  osivx_free(movie);
  return 0;
}


osivx_int
osivx_movie_close(
  osivx_movie_t * movie)
{
  switch (movie->type)
    {
    case OSIV_MOVIE_STREAM:
      return osivx_movie_closestream(movie);
    case OSIV_MOVIE_FILE:
      return osivx_movie_closefile(movie);
    case OSIV_MOVIE_TIFF_MERGE:
      return osivx_movie_closetiffmerge(movie);
    case OSIV_MOVIE_TIFF_SPLIT:
      return osivx_movie_closetiffsplit(movie);
    default:
      return OSIVX_EMOVTYPE;
    }
}

