/* FILE: osivx.c 
 * 
 * Copyright (c) 2002 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"


/**********************************************************
 * The following function traverses the passed arguments
 * and returns a non-zero error code if they are invalid.
 * The algorithm specific error code must still be called
 * in order to verify that the arguments are vaild, this
 * only checks the common limits.
 **********************************************************/
static osivx_int
osivx_shared_checkargs(osivx_args_t * a)
{
   if (a->alg_pproc < 0 ||
       a->alg_pproc > 5) {
      return OSIVX_EPREPROC;
   }

   if (a->alg_interp < 0 ||
       a->alg_interp > 4) {
      return OSIVX_EINTERP;
   }

   if (a->image_xsize < 1) {
      return OSIVX_EIMAGEXSIZE;
   }
   if (a->image_xsize > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->image_ysize < 1) {
      return OSIVX_EIMAGEYSIZE;
   }
   if (a->image_ysize > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->grid_xstart < 0) {
      return OSIVX_EGRIDXSTART;
   }
   if (a->grid_xstart > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->grid_ystart < 0) {
      return OSIVX_EGRIDYSTART;
   }
   if (a->grid_ystart > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->grid_xspacing < 1) {
      return OSIVX_EGRIDXSPACING;
   }
   if (a->grid_xspacing > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->grid_yspacing < 1) {
      return OSIVX_EGRIDYSPACING;
   }
   if (a->grid_yspacing > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->grid_xsize < 1) {
      return OSIVX_EGRIDXSIZE;
   }
   if (a->grid_xsize > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->grid_ysize < 1) {
      return OSIVX_EGRIDYSIZE;
   }
   if (a->grid_ysize > OSIVX_BIG_FRAME) {
      return OSIVX_EFRAMEINT;
   }

   if (a->wind_xsize < 1) {
      return OSIVX_EWINDXSIZE;
   }
   if (a->wind_xsize > OSIVX_BIG_WIND) {
      return OSIVX_EWINDINT;
   }

   if (a->wind_ysize < 1) {
      return OSIVX_EWINDYSIZE;
   }
   if (a->wind_ysize > OSIVX_BIG_WIND) {
      return OSIVX_EWINDINT;
   }

   if (a->wind_xsuper < a->wind_xsize) {
      return OSIVX_EWINDXSUPER;
   }
   if (a->wind_xsuper > OSIVX_BIG_WIND) {
      return OSIVX_EWINDINT;
   }

   if (a->wind_ysuper < a->wind_ysize) {
      return OSIVX_EWINDYSUPER;
   }
   if (a->wind_ysuper > OSIVX_BIG_WIND) {
      return OSIVX_EWINDINT;
   }

   if (a->disp_xmax < 1) {
      return OSIVX_EDISPXMAX;
   }
   if (a->disp_xmax > OSIVX_BIG_FRAME) {
      return OSIVX_EDISPINT;
   }

   if (a->disp_ymax < 1) {
      return OSIVX_EDISPYMAX;
   }
   if (a->disp_ymax > OSIVX_BIG_FRAME) {
      return OSIVX_EDISPINT;
   }

   if (a->disp_xiter < 1) {
      return OSIVX_EDISPXITER;
   }
   if (a->disp_xiter > OSIVX_BIG_FRAME) {
      return OSIVX_EDISPINT;
   }

   if (a->disp_yiter < 1) {
      return OSIVX_EDISPYITER;
   }
   if (a->disp_yiter > OSIVX_BIG_FRAME) {
      return OSIVX_EDISPINT;
   }

   if (a->movie_start < 0) {
      return OSIVX_EMOVIESTART;
   }
   if (a->movie_start > OSIVX_BIG_MOVIE) {
      return OSIVX_EDISPINT;
   }

   if (a->movie_finish < a->movie_start) {
      return OSIVX_EMOVIEFINISH;
   }
   if (a->movie_finish > OSIVX_BIG_MOVIE) {
      return OSIVX_EMOVIEINT;
   }

   if (a->movie_skip < 1) {
      return OSIVX_EMOVIESKIP;
   }
   if (a->movie_skip > OSIVX_BIG_MOVIE) {
      return OSIVX_EMOVIEINT;
   }

   if (a->movie_set < 0) {
      return OSIVX_EMOVIESET;
   }
   if (a->movie_set > OSIVX_BIG_MOVIE) {
      return OSIVX_EMOVIEINT;
   }

   return 0;
}


/**********************************************************
 * This function is the primary entry point into the
 * library.  It checks arguments, prepares files,
 * grids, etc.  Then passes control to algorithm code.
 * It returns zero on success and a non-zero code on
 * failure.
 **********************************************************/
osivx_int
osivx_run(
   osivx_args_t * a)
{
   osivx_opts_t *p = NULL;
   osivx_int scode = 0;

   /* allocate options struct */
   p = (osivx_opts_t *) osivx_calloc(1,sizeof(osivx_opts_t));
   if (p == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* check that arguments are tenable */
   scode = osivx_shared_checkargs(a);
   if (scode) goto finished;

   switch (a->alg_xcorr)
   {
   case 1:
      scode = osivx_alg1_checkargs(a);
      if (scode) goto finished;
      break;
   case 2:
      scode = osivx_alg2_checkargs(a);
      if (scode) goto finished;
      break;
   case 3:
      scode = osivx_alg3_checkargs(a);
      if (scode) goto finished;
      break;
   case 4:
      scode = osivx_alg4_checkargs(a);
      if (scode) goto finished;
      break;
   case 5:
      scode = osivx_alg5_checkargs(a);
      if (scode) goto finished;
      break;
   case 6:
      scode = osivx_alg6_checkargs(a);
      if (scode) goto finished;
      break;
   case 7:
      scode = osivx_alg7_checkargs(a);
      if (scode) goto finished;
      break;
   default:
      scode = OSIVX_EALG;
      goto finished;
   }

   /* keep handle to error info */
   p->err_info = &a->err_info;
   p->err_info->type = OSIVX_ERR_NONE;

   /* open up output file */
   if (a->vectors_name == NULL) {
      if (a->porthole_file == NULL) {
         scode = OSIVX_EVECFILE;
         goto finished;
      }

      p->vectors_file = osivx_fsopen(a->porthole_file, "w");
      if (p->vectors_file == NULL) {
         scode = OSIVX_EVECFILE;
         goto finished;
      }
   } else {
      if (a->porthole_file != NULL) {
        scode = OSIVX_EVECFILE;
        goto finished;
      }

      p->vectors_file = osivx_fopen(a->vectors_name, "w");
      if (!p->vectors_file) {
        scode = OSIVX_EVECFILE;
        goto finished;
      }
   }

   /* check that we have first movie file */
   if (a->movie_aname == NULL) {
      scode = OSIVX_EMOVAFILE;
      goto finished;
   }

   /* open up first movie file */
   p->movie_afile = osivx_movie_open(a->movie_aname,
     a->image_xsize, a->image_ysize, a->movie_start,
     a->movie_skip, a->movie_finish, &a->err_info);
   if (!p->movie_afile) {
      scode = OSIVX_EMOVAFILE;
      goto finished;
   }

   /* setup preprocessing for first movie */
   p->preproc_adata = osivx_preproc_init(
      p->movie_afile, a->alg_pproc);
   if (!p->preproc_adata) {
      scode = OSIVX_EMOVAFILE;
      goto finished;
   }

   /* check for optional second movie */
   if (a->movie_bname == NULL) {
      p->movie_bfile = p->movie_afile;
      p->preproc_bdata = p->preproc_adata;
   } else {
      /* open up second movie file */
      p->movie_bfile = osivx_movie_open(a->movie_bname,
        a->image_xsize, a->image_ysize, a->movie_start,
        a->movie_skip, a->movie_finish, &a->err_info);
      if (p->movie_bfile == NULL) {
         scode = OSIVX_EMOVBFILE;
	 goto finished;
      }

      /* setup preprocessing for second movie */
      p->preproc_bdata = osivx_preproc_init(
        p->movie_bfile, a->alg_pproc);
      if (p->preproc_bdata == NULL) {
         scode = OSIVX_EMOVBFILE;
	 goto finished;
      }
   }

   /* setup window grid engine */
   /* NOTE: for the moment we only have one grid,
    * type.  when this changes a switch statement
    * should be placed here to jump to the right
    * function
    */
   p->grid_data = osivx_grid_init_rect(a->grid_xstart,
     a->grid_ystart, a->grid_xspacing, a->grid_yspacing,
     a->grid_xsize, a->grid_ysize, a->wind_xsize,
     a->wind_ysize, a->wind_xsuper, a->wind_ysuper);
   if (p->grid_data == NULL) {
      scode = OSIVX_EGRID;
   }

   /* setup interpolation engine */
   p->interp_data = osivx_interp_init(a->alg_interp);
   if (p->interp_data == NULL) {
     scode = OSIVX_EINTERP;
     goto finished;
   }

   /* copy over variables */
   p->image_xsize = a->image_xsize;
   p->image_ysize = a->image_ysize;


   /* TODO: remove following once switch
    * to grid engine is finished
    */
   p->grid_xstart = a->grid_xstart;
   p->grid_ystart = a->grid_ystart;
   p->grid_xspacing = a->grid_xspacing;
   p->grid_yspacing = a->grid_yspacing;
   p->grid_xsize = a->grid_xsize;
   p->grid_ysize = a->grid_ysize;
   p->grid_len = p->grid_xsize * p->grid_ysize;
 
   /* TODO: remove following once switch
    * to grid engine is finished
    */
   p->wind_xsize = a->wind_xsize;
   p->wind_ysize = a->wind_ysize;
   p->wind_xsuper = a->wind_xsuper;
   p->wind_ysuper = a->wind_ysuper;
 
   p->disp_xoffset = a->disp_xoffset;
   p->disp_yoffset = a->disp_yoffset;
   p->disp_xmax = a->disp_xmax;
   p->disp_ymax = a->disp_ymax;
   p->disp_xiter = a->disp_xiter;
   p->disp_yiter = a->disp_yiter;

   p->movie_start = a->movie_start;
   p->movie_finish = a->movie_finish;
   p->movie_skip = a->movie_skip;
   p->movie_set = a->movie_set;
   p->num_iter = a->num_iter;
   p->num_thrd = a->num_thrd;
   p->write_map = (a->write_map != 0);
 
   /* calculate internal variables */
   p->image_len = p->image_xsize * p->image_ysize;
   p->wind_len = p->wind_xsize * p->wind_ysize;

   p->map_xsize = 2 * p->disp_xmax + 1;
   p->map_ysize = 2 * p->disp_ymax + 1;
   p->map_len = p->map_xsize * p->map_ysize;
 

   /* TODO: remove following once switch
    * to grid engine is finished
    */
   p->dom_xsize = (p->grid_xsize - 1) *
     p->grid_xspacing + p->wind_xsize + 1;
   p->dom_ysize = (p->grid_ysize - 1) *
     p->grid_yspacing + p->wind_ysize + 1;
   p->dom_len = p->dom_xsize * p->dom_ysize;

   p->fft_xsize = 2 * (p->wind_xsize / 2 + 1);
   p->fft_ysize = p->wind_ysize;
   p->fft_len = p->fft_xsize * p->fft_ysize;
 
   p->fft_xsuper = 2 * (p->wind_xsuper / 2 + 1);
   p->fft_ysuper = p->wind_ysuper;
   p->fft_slen = p->fft_xsuper * p->fft_ysuper;

   /* write parameters to file */
   scode = osivx_fwrite_header(p->vectors_file, a);
   if (scode) goto finished;

   /* run correlation */
   switch (a->alg_xcorr)
   {
   case 1: {
      osivx_wind_alg_t alg;
      osivx_alg1_make(&alg); 
      scode = osivx_generic_wind_alg(&alg, p);
      if (scode) goto finished;
   } break;
   case 2:
      scode = osivx_alg2(p);
      if (scode) goto finished;
      break;
   case 3: {
      osivx_wind_alg_t alg;
      osivx_alg3_make(&alg); 
      scode = osivx_generic_wind_alg(&alg, p);
      if (scode) goto finished;
   } break;
   case 4:
      scode = osivx_alg4(p);
      if (scode) goto finished;
      break;
   case 5: {
      osivx_wind_alg_t alg;
      osivx_alg5_make(&alg); 
      scode = osivx_generic_wind_alg(&alg, p);
      if (scode) goto finished;
   } break;
   case 6: {
      osivx_wind_alg_t alg;
      osivx_alg6_make(&alg);
      scode = osivx_generic_wind_alg(&alg, p);
      if (scode) goto finished;
   } break;
   case 7: {
      osivx_wind_alg_t alg;
      osivx_alg7_make(&alg);
      scode = osivx_generic_wind_alg(&alg, p);
      if (scode) goto finished;
   } break;
   default:
      scode = OSIVX_EALG;
      goto finished;
   }

finished:

   /* close first movie file*/
   if (p->movie_afile) {
      if (scode) {
	 osivx_movie_close(p->movie_afile);
      } else {
	 scode = osivx_movie_close(p->movie_afile);
      }
   }

   /* close second movie file */
   if (p->movie_bfile && p->movie_bfile != p->movie_afile) {
      if (scode) {
	 osivx_movie_close(p->movie_bfile);
      } else {
	 scode = osivx_movie_close(p->movie_bfile);
      }
   }

   /* close first preproc data */
   if (p->preproc_adata) {
      if (scode) {
	 osivx_preproc_free(p->preproc_adata);
      } else {
	 scode = osivx_preproc_free(p->preproc_adata);
      }
   }

   /* close second preproc data */
   if (p->preproc_bdata && p->preproc_bdata != p->preproc_adata) {
      if (scode) {
	 osivx_preproc_free(p->preproc_bdata);
      } else {
	 scode = osivx_preproc_free(p->preproc_bdata);
      }
   }

   /* free grid data */
   if (p->grid_data) {
      if (scode) {
        osivx_grid_free(p->grid_data);
      } else {
        scode = osivx_grid_free(p->grid_data);
      }
   }

   /* free interpolation data */
   if (p->interp_data) {
      if (scode) {
        osivx_interp_free(p->interp_data);
      } else {
        scode = osivx_interp_free(p->interp_data);
      }
   }

   /* close output file */
   if (p->vectors_file) {
      if (scode) {
	osivx_fclose(p->vectors_file);
      } else {
	scode = osivx_fclose(p->vectors_file);
      }
   }

   /* free memory */
   if (p) {
      osivx_free(p);
   }

   /* return status */
   return scode;
}
