/* FILE: preproc.c 
 *
 * Copyright (c) 2002-2006 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "osivx-devel.h"

#include <string.h>
#include <math.h>


/**********************************************************
 * The following function opens an osivx_preproc_t for
 * subtracting the minimum pixel in a sequence from
 * every image.  It returns a zero on success and 
 * an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_init1(
   osivx_preproc_t* preproc,
   osivx_movie_t* movie)
{
   osivx_int i;
   osivx_imcp_t *im = NULL;
   osivx_imcp_t *min = NULL;
   osivx_int scode = 0;

   /* allocate memory for temporary image
    * and for saved minimum image
    */
   im = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   min = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   if (im == NULL || min == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* use last image as start */
   scode = osivx_movie_read(movie, movie->lastf, im);
   if (scode) goto finished;
   memcpy(min, im, movie->imlen * sizeof(osivx_imcp_t));

   /* calculate minimum field */
   for (i = movie->firstf; i <= movie->lastf; i += movie->skipf) {
      osivx_int j;

      scode = osivx_movie_read(movie, i, im);
      if (scode) goto finished;

      for (j = 0; j < movie->imlen; j++) {
         if (*(im + j) < *(min + j)) {
            *(min + j) = *(im + j);
         }
      }
   }

   preproc->data = (void *) min;

finished:

   if (im != NULL) {
      osivx_free(im);
   }

   if (scode) {
      if (min != NULL) {
         osivx_free(min);
      }
   }

   return scode;
}


/**********************************************************
 * The following function opens an osivx_preproc_t for
 * adding up to the maximum pixel in a sequence from
 * every image.  It returns a zero on success and 
 * an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_init2(
   osivx_preproc_t * preproc,
   osivx_movie_t * movie)
{
   osivx_int i;
   osivx_imcp_t *im = NULL;
   osivx_imcp_t *max = NULL;
   osivx_int scode = 0;

   /* allocate memory for temporary image
    * and for saved minimum image
    */
   im = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   max = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   if (im == NULL || max == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* use last image as start */
   scode = osivx_movie_read(movie, movie->lastf, im);
   if (scode) goto finished;
   memcpy(max, im, movie->imlen * sizeof(osivx_imcp_t));

   /* calculate maximum field */
   for (i = movie->firstf; i <= movie->lastf; i += movie->skipf) {
      osivx_int j;

      scode = osivx_movie_read(movie, i, im);
      if (scode) goto finished;

      for (j = 0; j < movie->imlen; j++) {
         if (*(im + j) > *(max + j)) {
            *(max + j) = *(im + j);
         }
      }
   }

   preproc->data = (void *) max;

finished:

   if (im != NULL) {
      osivx_free(im);
   }

   if (scode) {
      if (max != NULL) {
         osivx_free(max);
      }
   }

   return scode;
}


/**********************************************************
 * The following function opens an osivx_preproc_t for
 * stretching an image to the minimum and minimum pixel
 * value in a sequence.  It returns a zero on success and 
 * an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_init3(
   osivx_preproc_t * preproc,
   osivx_movie_t * movie)
{
   osivx_int i;
   osivx_imcp_t *im = NULL;
   osivx_imcp_t *min = NULL;
   osivx_imcp_t *max = NULL;
   osivx_int scode = 0;

   /* allocate memory for temporary image
    * and for saved minimum/maximum images
    */
   im = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   min = (osivx_imcp_t *)osivx_malloc(
      2 * movie->imlen * sizeof(osivx_imcp_t));
   max = min + movie->imlen;
   if (im == NULL || min == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* use last image as start */
   scode = osivx_movie_read(movie, movie->lastf, im);
   if (scode) goto finished;
   memcpy(min, im, movie->imlen * sizeof(osivx_imcp_t));
   memcpy(max, im, movie->imlen * sizeof(osivx_imcp_t));

   /* calculate minimum/maximum field */
   for (i = movie->firstf; i <= movie->lastf; i += movie->skipf) {
      osivx_int j;

      scode = osivx_movie_read(movie, i, im);
      if (scode) goto finished;

      for (j = 0; j < movie->imlen; j++) {
         if (*(im + j) < *(min + j)) {
            *(min + j) = *(im + j);
         }
         if (*(im + j) > *(max + j)) {
            *(max + j) = *(im + j);
         }
      }
   }

   preproc->data = (void *) min;

finished:

   if (im != NULL) {
      osivx_free(im);
   }

   if (scode) {
      if (min != NULL) {
         osivx_free(min);
      }
   }

   return scode;
}


/**********************************************************
 * The following function opens an osivx_preproc_t for
 * stretching an image pixel values according to the
 * mean and standard deviation of the pixel values in
 * a sequence.  It returns a zero on success and an
 * error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_init4(
   osivx_preproc_t * preproc,
   osivx_movie_t * movie)
{
   osivx_int i;
   osivx_int j;
   osivx_imcp_t *im = NULL;
   osivx_real *ex = NULL;
   osivx_real *ex2 = NULL;
   osivx_int scode = 0;

   /* allocate memory for temporary image
    * and for saved E[X]/E[X^2] images
    */
   im = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   ex = (osivx_real *) osivx_calloc(
      2 * movie->imlen, sizeof(osivx_real));
   ex2 = ex + movie->imlen;
   if (im == NULL || ex == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* calculate stats sum field */
   for (i = movie->firstf; i <= movie->lastf; i += movie->skipf) {
      scode = osivx_movie_read(movie, i, im);
      if (scode) goto finished;

      for (j = 0; j < movie->imlen; j++) {
         *(ex + j) += *(im + j);
         *(ex2 + j) += ((osivx_real)*(im + j)) * *(im + j);
      }
   }

   /* convert sum to expectation */
   i = (movie->lastf - movie->firstf) / movie->skipf + 1;
   for (j = 0; j < movie->imlen; j++) {
      *(ex + j) /= i;
      *(ex2 + j) /= i;
      *(ex2 + j) -= *(ex + j) * *(ex + j);
      *(ex2 + j) = sqrt(*(ex2 + j));
   }

   preproc->data = (void *) ex;

finished:

   if (im != NULL) {
      osivx_free(im);
   }

   if (scode) {
      if (ex != NULL) {
         osivx_free(ex);
      }
   }

   return scode;
}


/**********************************************************
 * The following function opens an osivx_preproc_t for
 * stretching an image pixel values according to the
 * mean and standard deviation of the pixel values in
 * a sequence.  It returns a zero on success and an
 * error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_init5(
   osivx_preproc_t * preproc,
   osivx_movie_t * movie)
{
   osivx_int i;
   osivx_int j;
   osivx_imcp_t *im = NULL;
   osivx_real *avg = NULL;
   osivx_int scode = 0;

   /* allocate memory for temporary image
    * and for saved E[X] image
    */
   im = (osivx_imcp_t *)osivx_malloc(
      movie->imlen * sizeof(osivx_imcp_t));
   avg = (osivx_real *)osivx_calloc(
      movie->imlen, sizeof(osivx_real));
   if (im == NULL || avg == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   /* calculate sum field */
   for (i = movie->firstf; i <= movie->lastf; i += movie->skipf) {
      scode = osivx_movie_read(movie, i, im);
      if (scode) goto finished;

      for (j = 0; j < movie->imlen; j++) {
         *(avg + j) += *(im + j);
      }
   }

   /* calculate expectation from sum */
   i = (movie->lastf - movie->firstf) / movie->skipf + 1;
   for (j = 0; j < movie->imlen; j++) {
      *(avg + j) /= i;
   }

   preproc->data = (void *)avg;

finished:

   if (im != NULL) {
      osivx_free(im);
   }

   if (scode) {
      if (avg != NULL) {
         osivx_free(avg);
      }
   }

   return scode;
}


/**********************************************************
 * The following function opens an osivx_preproc_t for
 * the given movie sequence, using the specified
 * method.  It returns a zero on success and an
 * error code on failure.
 **********************************************************/
osivx_preproc_t *
osivx_preproc_init(
   osivx_movie_t * movie,
   osivx_int method)
{
   osivx_preproc_t *preproc = NULL;
   osivx_int scode = 0;

   preproc = (osivx_preproc_t *)
     osivx_malloc(sizeof(osivx_preproc_t));
   if (preproc == NULL) {
      scode = OSIVX_EMEM;
      goto finished;
   }

   preproc->movie = movie;
   preproc->type = method;

   switch (method) {
   case 0:
      break;
   case 1:
      scode = osivx_preproc_init1(preproc, movie);
      break;
   case 2:
      scode = osivx_preproc_init2(preproc, movie);
      break;
   case 3:
      scode = osivx_preproc_init3(preproc, movie);
      break;
   case 4:
      scode = osivx_preproc_init4(preproc, movie);
      break;
   case 5:
      scode = osivx_preproc_init5(preproc, movie);
      break;
   default:
      scode = OSIVX_EPREPROC;
   }

finished:
   if (scode) {
      if (preproc != NULL) {
         osivx_free(preproc);
         preproc = NULL;
      }
   }

   return preproc;
}


/**********************************************************
 * The following function executes "subtract to minima"
 * preprocessing on the given frame.  It returns a zero
 * on success and an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_exec1(
   osivx_preproc_t * preproc,
   osivx_int frm,
   osivx_imcp_t * im)
{
   osivx_int i;
   osivx_imcp_t *min;

   min = (osivx_imcp_t *) preproc->data;

   i = preproc->imlen - 1;
   for (; i > 0; i--) {
      if (*(im + i) < *(min + i)) {
         *(im + i) = 0;
      } else {
         *(im + i) -= *(min + i);
      }
   }

   return 0;
}


/**********************************************************
 * The following function executes "add to maxima"
 * preprocessing on the given frame.  It returns a zero
 * on success and an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_exec2(
   osivx_preproc_t * preproc,
   osivx_int frm,
   osivx_imcp_t * im)
{
   osivx_int i;
   osivx_imcp_t *max;

   max = (osivx_imcp_t *) preproc->data;

   i = preproc->imlen - 1;
   for (; i > 0; i--) {
      if (*(im + i) > *(max + i)) {
         *(im + i) = OSIVX_IMCP_MAX;
      } else {
         *(im + i) += OSIVX_IMCP_MAX - *(max + i);
      }
   }

   return 0;
}


/**********************************************************
 * The following function executes "stretch to limits"
 * preprocessing on the given frame.  It returns a zero
 * on success and an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_exec3(
   osivx_preproc_t * preproc,
   osivx_int frm,
   osivx_imcp_t * im)
{
   osivx_int i;
   osivx_imcp_t *min;
   osivx_imcp_t *max;

   min = (osivx_imcp_t *) preproc->data;
   max = min + preproc->imlen;

   i = preproc->imlen - 1;
   for (; i > 0; i--) {
      register osivx_real c = *(im + i);

      c -= *(min + i);
      c *= OSIVX_IMCP_MAX;
      c /= *(max + i) - *(min + i);
      if (c < 0) {
         *(im + i) = 0;
      } else if (c > OSIVX_IMCP_MAX) {
         *(im + i) = OSIVX_IMCP_MAX;
      } else {
         *(im + i) = c;
      }
   }

   return 0;
}


/**********************************************************
 * The following function executes "stretch to variance"
 * preprocessing on the given frame.  It returns a zero
 * on success and an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_exec4(
   osivx_preproc_t * preproc,
   osivx_int frm,
   osivx_imcp_t * im)
{
   osivx_int i;
   osivx_real *avg;
   osivx_real *std;

   avg = (osivx_real *) preproc->data;
   std = avg + preproc->imlen;

   i = preproc->imlen - 1;
   for (; i > 0; i--) {
      register osivx_real c = *(im + i);

      c -= *(avg + i) - 2 ** (std + i);
      c *= OSIVX_IMCP_MAX / (4 ** (std + i));
      if (c < 0) {
         *(im + i) = 0;
      } else if (c > OSIVX_IMCP_MAX) {
         *(im + i) = OSIVX_IMCP_MAX;
      } else {
         *(im + i) = c;
      }
   }

   return 0;
}


/**********************************************************
 * The following function executes "stretch to average"
 * preprocessing on the given frame.  It returns a zero
 * on success and an error code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_exec5(
   osivx_preproc_t * preproc,
   osivx_int frm,
   osivx_imcp_t * im)
{
   osivx_int i;
   osivx_real *avg;

   avg = (osivx_real *) preproc->data;

   i = preproc->imlen - 1;
   for (; i > 0; i--) {
      register osivx_real c = *(im + i);

      c *= OSIVX_IMCP_MAX / (2 * *(avg + i));
      *(im + i) = c < OSIVX_IMCP_MAX ? c : OSIVX_IMCP_MAX;
   }

   return 0;
}


/**********************************************************
 * The following function executes the preprocessing on
 * the given frame.  It returns a zero on success and
 * an error code on failure.
 **********************************************************/
osivx_int
osivx_preproc_exec(
   osivx_preproc_t * preproc,
   osivx_int frm,
   osivx_imcp_t * im)
{
   switch (preproc->type) {
   case 0:
      return 0;
   case 1:
      return osivx_preproc_exec1(preproc, frm, im);
   case 2:
      return osivx_preproc_exec2(preproc, frm, im);
   case 3:
      return osivx_preproc_exec3(preproc, frm, im);
   case 4:
      return osivx_preproc_exec4(preproc, frm, im);
   case 5:
      return osivx_preproc_exec5(preproc, frm, im);
   default:
      return OSIVX_EPREPROC;
   }
}


/**********************************************************
 * The following function frees the given preprocessing
 * algorithm. It returns a zero on success and an error
 * code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_free1(
   osivx_preproc_t * preproc)
{
   osivx_free(preproc->data);
   osivx_free(preproc);
   return 0;
}


/**********************************************************
 * The following function frees the given preprocessing
 * algorithm. It returns a zero on success and an error
 * code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_free2(
   osivx_preproc_t * preproc)
{
   osivx_free(preproc->data);
   osivx_free(preproc);
   return 0;
}


/**********************************************************
 * The following function frees the given preprocessing
 * algorithm. It returns a zero on success and an error
 * code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_free3(
   osivx_preproc_t * preproc)
{
   osivx_free(preproc->data);
   osivx_free(preproc);
   return 0;
}


/**********************************************************
 * The following function frees the given preprocessing
 * algorithm. It returns a zero on success and an error
 * code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_free4(
   osivx_preproc_t * preproc)
{
   osivx_free(preproc->data);
   osivx_free(preproc);
   return 0;
}


/**********************************************************
 * The following function frees the given preprocessing
 * algorithm. It returns a zero on success and an error
 * code on failure.
 **********************************************************/
static osivx_int
osivx_preproc_free5(
   osivx_preproc_t * preproc)
{
   osivx_free(preproc->data);
   osivx_free(preproc);
   return 0;
}


/**********************************************************
 * The following function frees the given preprocessing
 * engine.  It returns a zero on success and an error
 * code on failure.
 **********************************************************/
osivx_int
osivx_preproc_free(
   osivx_preproc_t * preproc)
{
   switch (preproc->type) {
   case 0:
      osivx_free(preproc);
      return 0;
   case 1:
      return osivx_preproc_free1(preproc);
   case 2:
      return osivx_preproc_free2(preproc);
   case 3:
      return osivx_preproc_free3(preproc);
   case 4:
      return osivx_preproc_free4(preproc);
   case 5:
      return osivx_preproc_free5(preproc);
   default:
      return OSIVX_EPREPROC;
   }
}
