/* FILE: zfile.c 
 * 
 * Copyright (c) 2002 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */

#include "zargs.h"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>


#define BUFLEN 1024

static void
zfile_xcomment(
  char* s)
{
   char *c;

   c = strchr(s, '#');
   while (c != NULL)
   {
      if ( c == s ||
         *(c-1)!='\\')
      {
         *c = '\0';
         return;
      }     

      c = strchr(c+1,'#');
   }
}

static void
zfile_xleading(
  char* s)
{
   size_t pad, len;

   pad = strspn(s, " \t\n");
   len = strlen(s+pad) + 1;
   memmove(s, s+pad, len);
}

static void
zfile_xtrailing(
  char* s)
{
   size_t d, p;

   do {
      p = strspn(s," \t\n");
      d = strcspn(s+p," \t\n");
      if (d == 0) break;
      else s += p + d;
   } while (1);
   
   *s = '\0';  
}

static int
zfile_isjunk(
   const char* s)
{
   while (*s != '\0') {
      if (!isprint(*s))
         return 1;

      s++;
   }

   return 0;
}

static int
zfile_split(
   char* var,
   char** val)
{
   char *c;

   c = strchr(var, '=');
   if (c == NULL) {
      return 1;
   }

   *val = c+1; *c++ = '\0';
   zfile_xtrailing(var); 
   zfile_xleading(*val); 

   return 0;
}

static int
zfile_comp(const void* a, const void* b)
{
   zfile_item *ax = (zfile_item*)a;
   zfile_item *bx = (zfile_item*)b;
   return strcmp(ax->var,bx->var);
}

static zfile_item*
zfile_search(
   zfile_item* list,
   unsigned nlist,
   const char* var)
{
    zfile_item key, *fnd;

    /* set key */
    key.var = var;
    key.evt = 0;

    /* do search */ 
    fnd = bsearch(&key, list, nlist,
      sizeof(zfile_item), zfile_comp);

    /* return item */
    return fnd;
}

int
zfile_verify(
   zfile_item* l,
   unsigned n)
{
   unsigned i;

   /* check first exists */
   if (l->var == NULL) {
      return 0;
   }

   /* traverse list */
   for (i = 1; i < n; i++)
   {
      /* check total length */
      if ((l+i)->var == NULL)
         return i;

      /* check item order */
      if (zfile_comp(l+i-1,l+i)>=0)
         return i;
   }

   /* check that list is over */
   if ((l+i)->var != NULL)
      return i;

   /* return success */
   return -1;
}

int
zfile_parse(
   const char* fn,
   zfile_item* list,
   unsigned nlist,
   zfile_callb cb,
   void* ptr)
{
   char line[BUFLEN];
   char *cc, *sv;
   FILE* fp; 
  
   fp = fopen(fn, "r");
   if (fp == NULL)
   {
      (*cb)(ZFILE_EOPEN,fn,
          NULL, ptr);
      return ZFILE_EOPEN;
   }

   cc = fgets(line, BUFLEN, fp);
   while (cc != NULL)
   {
      if (strlen(line)==BUFLEN)
      {
         (*cb)(ZFILE_EBUFF,line,
             NULL, ptr);
         return ZFILE_EBUFF;
      }

      zfile_xcomment(line);
      zfile_xleading(line);
      zfile_xtrailing(line);

      if (*line != '\0')
      {
         zfile_item* it;

         if (zfile_isjunk(line))
         {
            (*cb)(ZFILE_EJUNK,NULL,
               NULL, ptr);
            return ZFILE_EJUNK;

         }

         if (zfile_split(line, &sv))
         {
            (*cb)(ZFILE_EFORM,line,
               NULL, ptr);
            return ZFILE_EFORM;
         }

         it = zfile_search(list,
            nlist, line);

         if (it == NULL)
         {
            (*cb)(ZFILE_ELOST,line,
                NULL, ptr);
            return ZFILE_ELOST;
         }

         (*cb)(it->evt, line, sv, ptr);
      }

      cc = fgets(line, BUFLEN, fp);
   }

   return 0;
}
