/* FILE: zopts.c
 * 
 * Copyright (c) 2002 James A. Strother 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 * 
 */


#include "zargs.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define BUFLEN 1024

static int
zopts_isopt(const char* c)
{
   if (*c == '-' && *(c+1) != '\0')
      return 1;

   return 0;
}

static int
zopts_comp(const void* a, const void* b)
{
   zopts_item *ax = (zopts_item*)a;
   zopts_item *bx = (zopts_item*)b;
   return strcmp(ax->opt,bx->opt);
}

static zopts_item*
zopts_search(
   zopts_item* list,
   unsigned nlist,
   const char* opt)
{
    zopts_item key, *fnd;

    /* prepare key */
    key.opt = opt;
    key.type = 0;
    key.evt = 0;

    /* do search */
    fnd = bsearch(&key, list, nlist,
      sizeof(zopts_item), zopts_comp);

    /* return item */
    return fnd;
}

int
zopts_verify(
   zopts_item* l,
   unsigned n)
{
    unsigned i;

    /* check first exists */
    if (l->opt == NULL)
	return 0;

    /* traverse list */
    for (i = 1; i < n; i++)
    {
	/* check total length */
	if ((l+i)->opt == NULL)
	    return i;

	/* check item order */
	if (zopts_comp(l+i-1,l+i)>=0)
	    return i;
    }

    /* check that list is over */
    if ((l+i)->opt != NULL)
	return i;

    /* return success */
    return -1;
}

int
zopts_parse(
   int argc,
   char* argv[],
   zopts_item* long_list,
   unsigned long_nlist,
   zopts_item* shrt_list,
   unsigned shrt_nlist,
   zopts_callb cb,
   void* ptr) 
{
    char c[BUFLEN];
    char m[BUFLEN];
    unsigned i;

    /* traverse the arguments */
    for (i = 1; i < argc; i++)
    {
	zopts_item* item;
	char* ac = NULL;

        /* check arg length */
        if (strlen(argv[i])>=BUFLEN)
	{
	    (*cb)(ZOPTS_EBUFF,
               argv[i],NULL,ptr);
	    return ZOPTS_EBUFF;
	}

	/* copy string to buffer */
	strcpy(c, argv[i]);

	/* check whether is option */
	if (!zopts_isopt(c))
	    return i;

	/* check for end of options */
	if (!strcmp(c, "--"))
	    return i+1;

        /* check for long/short opts */
	if (!strncmp(c,"--",2))
	{
            /* find the argument */
	    ac = strchr(c, '=');
	    if (ac != NULL) {
		*ac++ = '\0';
	    }

	    /* find item in long list */
	    item = zopts_search(
               long_list,long_nlist,c+2);
	} else {
            /* zopt_isopt() ensures it is
               short option with at least
               three characters
            */
            if (c[2] != '\0') {
               strcpy(m, c + 2);
               ac = m;
               c[2] = '\0';
            } else {
               ac = NULL;
            }

            /* find item in short list */
            item = zopts_search(
               shrt_list,shrt_nlist,c+1);
        }

	/* check if item found */
	if (item == NULL)
        {
	    (*cb)(ZOPTS_ELOST, c,
                NULL, ptr);
	    return ZOPTS_ELOST;
	}

	/* search for argument */
        switch (item->type) {
	case ZOPTS_NARG:
	    if (ac != NULL)
	    {
		(*cb)(ZOPTS_EWONT, c,
                    NULL, ptr);
		return ZOPTS_EWONT;
	    }
            break;
        case ZOPTS_WARG:
	    if (ac == NULL)
	    {
		if (i == argc-1)
		{
		    (*cb)(ZOPTS_ENEED, c,
                        NULL, ptr);
		    return ZOPTS_ENEED;
		}
		
		ac = argv[i++ + 1];
		if (zopts_isopt(ac))
		{
		    (*cb)(ZOPTS_ENEED, c,
                        NULL, ptr);
		    return ZOPTS_ENEED;
		}
	    }
            break;
        case ZOPTS_XARG:
	    if (ac == NULL && i<argc-1)
	    {
		ac = argv[i + 1];
		if (!zopts_isopt(ac)) i++;
		else ac = NULL;
	    }
            break;
        }
	
	/* perform callback */
	(*cb)(item->evt, c, ac, ptr);
    }

    return i;
}


